"""Import everything that should be available when using 'import topotoolbox'
"""
from .interface import *
from .grid_object import GridObject
from .flow_object import FlowObject
from .stream_object import StreamObject
from .graphflood_object import GFObject
from .graphflood import *
from .utils import *
from .stream_functions import *
from .swath import *



import os
import sys

import sphinx_book_theme

import topotoolbox


extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx.ext.todo',
    'nbsphinx',
    'myst_parser',
    'matplotlib.sphinxext.plot_directive'
]

project = 'TopoToolbox'
copyright = '2024, TopoToolbox Team'
author = 'TopoToolbox Team'
release = '3.0.1'

# -- General configuration ---------------------------------------------------

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', 'README.md']

# -- Options for HTML output -------------------------------------------------

html_theme = 'sphinx_book_theme'
html_static_path = ['_static']
html_logo = 'logo.png'

html_context = {
    "default_mode": "light",
}

html_theme_options = {
    "repository_url": "https://github.com/TopoToolbox/pytopotoolbox",
    "use_repository_button": True,
}

# -- Options for nbsphinx ----------------------------------------------------

nbsphinx_allow_errors = True

# -- Options for autodoc -----------------------------------------------------

autodoc_default_options = {
    'members': True,
    'undoc-members': True,
    'private-members': False,
    'inherited-members': True,
    'show-inheritance': True,
}

autosummary_generate = True  # Enable autosummary to generate stub files

# -- Options for nbgallery ---------------------------------------------------

nbsphinx_thumbnails = {
    'examples/downloading': '_static/thumbnails/placeholder.png',
    'examples/flowobject': '_static/thumbnails/placeholder.png',
    'examples/streamobject': '_static/thumbnails/placeholder.png',
}

# -- Options for plot_directive

plot_include_source = True



"""This module contains the StreamObject class.
"""
import math
import warnings
import copy

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
from scipy.sparse import csr_matrix
from shapely.geometry import LineString
import geopandas as gpd
import scipy.sparse as sp

# pylint: disable=no-name-in-module
from clarabel import ZeroConeT, NonnegativeConeT, DefaultSettings, DefaultSolver

from .flow_object import FlowObject
from .grid_object import GridObject
from .interface import validate_alignment
from .stream_functions import imposemin

# pylint: disable=no-name-in-module
from . import _stream  # type: ignore

_all_ = ['StreamObject']


class StreamObject():
    """A class to represent stream flow accumulation based on a FlowObject.
    """

    def __init__(self, flow: FlowObject, units: str = 'pixels',
                 threshold: int | float | GridObject | np.ndarray = 0,
                 stream_pixels: GridObject | np.ndarray | None = None,
                 channelheads: np.ndarray | None = None
                 ) -> None:
        """Initializes the StreamObject by processing flow accumulation.

        Parameters
        ----------
        flow : FlowObject
          The input flow object containing source, target, direction, and other
          properties related to flow data.
        units : str, optional
          Units of measurement for the flow data. Can be 'pixels', 'mapunits',
          'm2', or 'km2'. Default is 'pixels'.
        threshold : int | float | GridObject | np.ndarray, optional
          The upslope area threshold for flow accumulation. This can be an
          integer, float, GridObject, or a NumPy array. If more water than in
          the threshold has accumulated in a cell, it is part of the stream.
          Default is 0, which will result in the threshold being generated
          based on this formula: threshold = (avg^2)*0.01
          where shape = (n,m).
        stream_pixels : GridObject | np.ndarray, optional
          A GridObject or np.ndarray made up of zeros and ones to denote where
          the stream is located. Using this will overwrite any use of the
          threshold argument.
        channelheads: (rows, cols), optional
          A tuple of two array-like objects containing the row and
          column indices of the channel heads. All streams downstream of
          the indicated channel heads will be returned in the
          StreamObject.

        Raises
        ------
        ValueError
          If the `units` parameter is not 'pixels', 'mapunits', 'm2', or 'km2'.
        ValueError
          If the shape of the threshold does not match the flow object shape.

        Example
        -------
        .. plot ::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('perfectworld')
           >>> fd = topotoolbox.FlowObject(dem)
           >>> s = topotoolbox.StreamObject(fd,threshold=1000,units='pixels')
           >>> fig, ax = plt.subplots()
           >>> _= dem.plot(cmap="terrain")
           >>> _= s.plot(color='r')
           >>> plt.show()
        """
        if not isinstance(flow, FlowObject):
            err = f"{flow} is not a topotoolbox.FlowObject."
            raise TypeError(err)

        self.cellsize = flow.cellsize
        self.shape = flow.shape
        self.strides = flow.strides
        self._order = flow.order

        # georeference
        self.bounds = flow.bounds
        self.transform = flow.transform
        self.georef = flow.georef

        cell_area = 0.0
        # Calculate the are of a cell based on the units argument.
        if units == 'pixels':
            cell_area = 1.0
        elif units == 'm2':
            cell_area = self.cellsize**2
        elif units == 'km2':
            cell_area = (self.cellsize*0.001)**2
        elif units == 'mapunits':
            if self.georef is not None:
                if self.georef.is_projected:
                    # True so cellsize is in meters
                    cell_area = self.cellsize**2
                else:
                    # False so cellsize is in degrees
                    pass
        else:
            err = (f"Invalid unit '{units}' provided. Expected one of "
                   f"'pixels', 'mapunits', 'm2', 'km2'.")
            raise ValueError(err) from None

        # If stream_pixels are provided, the stream can be generated based
        # on stream_pixels without the need for a threshold
        w = np.zeros(flow.shape, dtype='bool', order=self._order).ravel(order='K')
        if stream_pixels is not None:
            if not validate_alignment(self, stream_pixels):
                err = (
                    f"stream_pixels shape {stream_pixels.shape}"
                    f" does not match FlowObject shape {self.shape}.")
                raise ValueError(err)

            w = (np.asarray(stream_pixels) != 0).ravel(order='K')

            if threshold != 0:
                warn = ("Since stream_pixels have been provided, the "
                        "input for threshold will be ignored.")
                warnings.warn(warn)
        elif channelheads is not None:
            ch = np.zeros(flow.shape, dtype=np.uint32, order=flow.order)
            ch[channelheads] = 1
            edges = np.ones(flow.source.size, dtype=np.uint32)
            _stream.traverse_down_u32_or_and(
                ch, edges, flow.source, flow.target)
            w = (ch > 0).ravel(order='K')

        # Create the appropriate threshold matrix based on the threshold input.
        else:
            if isinstance(threshold, (int, float)):
                if threshold == 0:
                    avg = (flow.shape[0] + flow.shape[1])//2
                    threshold = np.full(
                        self.shape, math.floor((avg ** 2) * 0.01),
                        order = self._order,
                        dtype=np.float32)
                else:
                    threshold = np.full(
                        self.shape, threshold, order=self._order, dtype=np.float32)
            else:
                if not validate_alignment(self, threshold):
                    err = (
                        f"Threshold GridObject shape {threshold.shape} does "
                        f"not match FlowObject shape: {self.shape}.")
                    raise ValueError(err) from None

                threshold = np.asarray(threshold, dtype=np.float32, order=self._order)

            # Divide the threshold by how many m^2 or km^2 are in a cell to
            # convert the user input to pixels for further computation.
            threshold /= cell_area

            # Generate the flow accumulation matrix (acc)
            acc = np.asarray(flow.flow_accumulation())

            # Generate a 1D array that holds all indexes where more water than
            # in the required threshold is collected. (acc >= threshold)
            w = (acc >= threshold).ravel(order='K')

        # Indices of pixels in the stream network
        # This is a node attribute list
        self.stream = np.nonzero(w)[0]

        # Find edges whose source pixel is in the stream network
        u = flow.source
        v = flow.target
        d = flow.direction.ravel(order='K')

        i = w[u]

        # Renumber the nodes of the stream network
        ix = np.zeros_like(w, dtype='int64')
        ix[w] = np.arange(0, self.stream.size)

        # Edges in the stream network
        #
        # Elements of these edge attribute lists are 0-based indices
        # into node attribute lists.
        #
        # To convert these to pixel indices in the original GridObject
        # or FlowObject, use stream[source] or stream[target].
        self.source = ix[u[i]]
        self.target = ix[v[i]]
        self.direction = d[u[i]]

        # misc
        self.path = flow.path
        self.name = flow.name

    @property
    def node_indices(self) -> tuple[np.ndarray, ...]:
        """The row and column indices of the nodes of the stream
        network.

        Returns
        -------
        tuple of ndarray
            A tuple of arrays containing the row indices and column
        indices for the nodes in the stream network. Each of these
        arrays is a node attribute lists and have a length equal to
        the number of nodes in the stream network. This tuple of
        arrays is suitable for indexing GridObjects or arrays shaped
        like the GridObject from which this StreamObject was derived.

        """
        return np.unravel_index(self.stream, self.shape, self._order)

    def node_indices_where(self, nal: np.ndarray) -> tuple[np.ndarray, ...]:
        """The row and column indices of the nodes of the stream
        network where a condition is satisfied.

        Returns
        -------
        tuple of ndarray
            A tuple of arrays containing the row indices and column
        indices for the nodes in the stream network where the input
        Boolean node attribute list `nal` is true. Each of these
        arrays is a node attribute lists and have a length equal to
        the number of nodes in the stream network. This tuple of
        arrays is suitable for indexing GridObjects or arrays shaped
        like the GridObject from which this StreamObject was derived.

        """
        return np.unravel_index(self.stream[nal], self.shape, self._order)

    @property
    def source_indices(self) -> tuple[np.ndarray, ...]:
        """The row and column indices of the sources of each edge in
        the stream network.

        Returns
        -------
        tuple of ndarray
            A tuple of arrays containing the row indices and column
        indices of the sources of each edge in the stream
        network. Each of these arrays is an edge attribute lists and
        have a length equal to the number of edges in the stream
        network. This tuple of arrays is suitable for indexing
        GridObjects or arrays shaped like the GridObject from which
        this StreamObject was derived.

        """
        return np.unravel_index(self.stream[self.source],
                                self.shape, self._order)

    @property
    def target_indices(self) -> tuple[np.ndarray, ...]:
        """The row and column indices of the targets of each edge in
        the stream network.

        Returns
        -------
        tuple of ndarray
            A tuple of arrays containing the row indices and column
        indices of the sources of each edge in the stream
        network. Each of these arrays is an edge attribute lists and
        have a length equal to the number of edges in the stream
        network. This tuple of arrays is suitable for indexing
        GridObjects or arrays shaped like the GridObject from which
        this StreamObject was derived.

        """
        return np.unravel_index(self.stream[self.target],
                                self.shape, self._order)

    @property
    def gridmask(self) -> np.ndarray:
        """mask with the shape of the original grid with stream
        indices.

        Returns
        -------
        ndarray
            2D array where rivers are 1 and non-river pixels
            are 0.

        """
        mask = np.zeros(self.shape, order = self._order, dtype = np.bool)
        mask[self.source_indices] = True
        mask[self.target_indices] = True
        return mask

    @property
    def coordinates(self) -> tuple[np.ndarray, np.ndarray]:
        """Return x and y coordinates of the nodes in the stream network.

        Returns
        -------
        tuple of ndarray
            A pair of node attribute lists containing x and y coordinates of each node.
        """
        # pylint: disable=unbalanced-tuple-unpacking
        j, i = self.node_indices
        return self.transform * (i, j)

    @property
    def x(self) -> np.ndarray:
        """Return x coordinates of the nodes in the stream network

        If you are trying to get both x and y coordinates, it is
        slightly more efficient to use `coordinates`.

        Returns
        -------
        ndarray
            A node attribute list containing the x coordinates of each node
        """
        return self.coordinates[0]

    @property
    def y(self) -> np.ndarray:
        """Return y coordinates of the nodes in the stream network

        If you are trying to get both x and y coordinates, it is
        slightly more efficient to use `coordinates`.

        Returns
        -------
        ndarray
            A node attribute list containing the y coordinates of each node

        """
        return self.coordinates[1]


    def node_to_node_distance(self) -> np.ndarray:
        """
        Compute the pixel-to-pixel distance for each edge.

        Returns
        -------
        np.ndarray, float32
            An edge attribute list with the distance between pixels

        Example
        -------
        >>> import topotoolbox
        >>> dem = topotoolbox.load_dem('perfectworld')
        >>> fd = topotoolbox.FlowObject(dem)
        >>> s = topotoolbox.StreamObject(fd,threshold=1000,units='pixels')
        >>> print(s.node_to_node_distance()) # doctest: +SKIP
        """
        d = np.abs(self.stream[self.source] - self.stream[self.target])

        dist = self.cellsize * np.where(
            (d == self.strides[0]) | (d == self.strides[1]),
            np.float32(1.0),
            np.sqrt(np.float32(2.0)))

        return dist

    def downstream_distance(self) -> np.ndarray:
        """Compute the maximum distance between a node in the stream
        network and the channel head.

        Returns
        -------
        np.ndarray, float32
            A node attribute list with the downstream distances
        """
        d = self.node_to_node_distance()  # Edge attribute list
        dds = np.zeros_like(self.stream, dtype=np.float32)
        _stream.traverse_down_f32_max_add(dds, d, self.source, self.target)

        return dds

    def upstream_distance(self) -> np.ndarray:
        """Compute the upstream distance for a node in the stream network

        Returns
        -------
        np.ndarray, float32
            A node attribute list with the upstream distances
        """
        d = self.node_to_node_distance()
        dds = np.zeros_like(self.stream, dtype=np.float32)
        _stream.traverse_up_f32_max_add(dds, d, self.source, self.target)

        return dds

    def distance(self, kind='from_outlet'):
        """Compute distances along the stream network

        Parameters
        ----------
        kind: string
           The kind of distance to return. One of

           - 'from_outlet' (default): distance in upstream direction
           - 'min_from_ch': shortest from channelhead
           - 'max_from_ch': longest from channelhead
           - 'node_to_node': distance between each node and its downstream neighbor

        Returns
        -------
        np.ndarray
           A node attribute list with the requested distances.

        Raises
        ------
        ValueError
            If method is currently unsupported
        """
        match kind:
            case 'from_outlet':
                return self.upstream_distance()
            case 'max_from_ch':
                return self.downstream_distance()
            case 'min_from_ch':
                d = self.node_to_node_distance()  # Edge attribute list
                dds = np.full_like(self.stream, np.inf, dtype=np.float32)
                dds[self.streampoi('channelheads')] = 0
                _stream.traverse_down_f32_min_add(dds, d, self.source, self.target)
                return dds
            case 'node_to_node':
                dds = np.zeros_like(self.stream, dtype=np.float32)
                dds[self.source] = self.node_to_node_distance()
                return dds
            case _:
                raise ValueError(f"""{kind} not supported. method should be one of
                'from_outlet', 'max_from_ch', 'min_from_ch', 'node_to_node'""")

    def ezgetnal(self, k, dtype=None):
        """Retrieve a node attribute list from k

        Parameters
        ----------
        k : GridObject | np.ndarray | float
            The object from which node values will be extracted. If
            `k` is a `GridObject` or an `np.ndarray` with the same
            shape as the underlying DEM of this `StreamObject`, the
            node values will be extracted from the grid by
            indexing. If `k` is an array with the same shape as the
            node attribute list, `ezgetnal` returns a copy of `k`. If
            `k` is a scalar value, `ezgetnal` returns an array of the
            right shape filled with `k`.

        Returns
        -------
        np.ndarray
            The resulting array will always be a copy of the input
            array.

        Raises
        ------
        ValueError
            If `k` does not have the right shape to be indexed by the
            `StreamObject`.
        """
        if np.isscalar(k):
            nal = np.full(self.stream.shape, k, dtype=None)
        else:
            if validate_alignment(self, k):
                # k is a GridObject or ndarray with the right shape
                # and georeferencing
                # Advanced indexing of k will always return a copy
                nal = k[self.node_indices]

                # We use copy=False in astype to avoid copying that copy
                # if possible
                nal = nal.astype(dtype or nal.dtype, copy=False)
            elif hasattr(k, "shape") and self.stream.shape == k.shape:
                # k is already a node attribute list
                nal = np.array(k, dtype=dtype, copy=True)
            else:
                raise ValueError(f"""{k} is not a node attribute list
                of the appropriate shape.""")

        return nal

    def streampoi(self, point_type: str) -> np.ndarray:
        """Extract points of interest from the stream network

        Currently supported points of interest are 'channelheads',
        'outlets' and 'confluences'

        Parameters
        ----------
        point_type: 'channelheads' or 'outlets' or 'confluences'
            The type of points to select from the stream network

        Returns
        -------
        np.ndarray
            A logical node attribute list indicating the locations of points.

        Raises
        ------
        ValueError
            If an unknown point type is requested.
        """
        indegree = np.zeros(self.stream.size, dtype=np.uint8)
        outdegree = np.zeros(self.stream.size, dtype=np.uint8)
        _stream.edgelist_degree(indegree, outdegree, self.source, self.target)
        if point_type == 'channelheads':
            return (outdegree > 0) & (indegree == 0)

        if point_type == 'outlets':
            return (outdegree == 0) & (indegree > 0)

        if point_type == 'confluences':
            return indegree > 1

        if point_type == 'bconfluences':
            confluences = indegree > 1
            edgesintoconfluences = confluences[self.target]
            bconfluences = np.zeros(self.stream.shape, dtype=bool)
            bconfluences[self.source] = edgesintoconfluences
            return bconfluences

        raise ValueError(f"{point_type} is not currently supported")

    def xy(self, data=None):
        """Compute the x and y coordinates of continuous stream segments

        Arguments
        ---------
        data: tuple, optional
           A tuple of two node attribute lists representing the
           desired x and y values for each pixel in the stream
           network. If this argument is not supplied, the returned x
           and y values are the geographic coordinates of the node.

        Returns
        -------
        list
            A list of lists of (x,y) pairs.
        """
        if data is None:
            # pylint: disable=unbalanced-tuple-unpacking
            j, i = self.node_indices
            xs, ys = self.transform * np.vstack((i, j))
        else:
            xs, ys = data

        vertices = range(self.stream.size)
        edges = range(self.source.size)

        # Construct an adjacency list for the graph,
        # so we can do a depth-first search
        adjacency_list = [[] for _ in vertices]
        for e in edges:
            src = self.source[e]
            tgt = self.target[e]
            adjacency_list[src].append(tgt)

        # Depth-first search of the graph
        visited = np.zeros(self.stream.size, dtype=bool)
        segments = []
        stack = []

        for e in edges:
            src = self.source[e]
            if not visited[src]:
                # Start a new segment
                stack.append(src)
                segments.append([])
            while stack:
                u = stack.pop()
                # Always append the next vertex to the segment
                segments[-1].append((xs[u], ys[u]))
                # If u has already been visited, we stop the segment
                # otherwise, we push its children to visit later
                if not visited[u]:
                    visited[u] = True
                    # Add any neighbors of
                    for v in adjacency_list[u]:
                        stack.append(v)

        return segments

    def to_geodataframe(self):
        '''Convert the stream network to a GeoDataFrame.

        Returns
        ----------
        geopandas.GeoDataFrame
            The GeoDataFrame.

        Example
        -------
        >>> import topotoolbox
        >>> import matplotlib.pyplot as plt
        >>> dem = topotoolbox.load_dem('bigtujunga')
        >>> fd = topotoolbox.FlowObject(dem)
        >>> s = topotoolbox.StreamObject(fd)
        >>> s_gdf = s.to_geodataframe()
        '''

        line_geoms = [LineString(coords) for coords in self.xy()]
        gdf = gpd.GeoDataFrame(geometry=line_geoms, crs=self.georef)
        return gdf

    def to_shapefile(self, path: str) -> None:
        '''Convert the stream network to a georeferenced shapefile.

        Parameters
        ----------
        path : str
             path where the shapefile will be saved.

        Example
        -------
        >>> import topotoolbox
        >>> import matplotlib.pyplot as plt
        >>> dem = topotoolbox.load_dem('bigtujunga')
        >>> fd = topotoolbox.FlowObject(dem)
        >>> s = topotoolbox.StreamObject(fd)
        >>> s.to_shapefile('stream_network.shp')
        '''
        gdf = self.to_geodataframe()
        gdf.to_file(path)

    def to_grid(self, nal=None):
        """Convert StreamObject to GridObject

        Parameters
        ----------
        nal : GridObject | np.ndarray | float, optional
            Node attribute list to map onto the grid. If None, creates a
            boolean grid with True at stream locations. If provided, creates
            a grid with the node values at stream locations and nan elsewhere.

        Returns
        -------
        GridObject
            A GridObject with the same spatial reference as the StreamObject.
            If nal is None, contains boolean values (True for streams).
            Otherwise, contains the nal values at stream locations and nan
            elsewhere.
        """
        # pylint: disable=import-outside-toplevel
        # Local import to avoid circular import
        from .utils import get_dtype

        if nal is None:
            # Create a boolean grid with True at stream locations
            z = self.gridmask
        else:

            # Create a grid filled with nan
            z = np.full(self.shape, np.nan, dtype=get_dtype(nal), order=self._order)

            # Set stream locations to the nal values
            z[self.node_indices] = nal[self.node_indices]

        # Create and return a GridObject
        grid = GridObject()
        grid.z = z
        grid.cellsize = self.cellsize
        grid.bounds = self.bounds
        grid.transform = self.transform
        grid.georef = self.georef
        grid.name = 'stream grid'

        return grid

    def plot(self, ax=None, scalex=True, scaley=True, **kwargs):
        """Plot the StreamObject

        Stream segments as computed by StreamObject.xy are plotted
        using a LineCollection.

        Parameters
        ----------
        ax: matplotlib.axes.Axes, optional
            The axes in which to plot the StreamObject. If no axes are
            given, the current axes are used.

        scalex: bool, optional
            Autoscale the x-axis limits. Defaults to `True`. If the
            x-axis limits have been set manually with `set_xlim`, the
            autoscaling will not be applied regardless of the value of
            `scalex`.

        scaley: bool, optional
            Autoscale the y-axis limits. Defaults to `True`. If the
            y-axis limits have been set manually with `set_ylim`, the
            autoscaling will not be applied regardless of the value of
            `scaley`.

        **kwargs
            Additional keyword arguments are forwarded to LineCollection

        Returns
        -------
        matplotlib.axes.Axes
            The axes into which the StreamObject has been plotted.

        Example
        -------
        .. plot ::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('perfectworld')
           >>> fd = topotoolbox.FlowObject(dem)
           >>> s = topotoolbox.StreamObject(fd,threshold=1000,units='pixels')
           >>> fig, ax = plt.subplots()
           >>> _= dem.plot(cmap="terrain")
           >>> _= s.plot(color='r')
           >>> plt.show()
        """

        if ax is None:
            ax = plt.gca()
        collection = LineCollection(self.xy(), **kwargs)
        ax.add_collection(collection)
        ax.autoscale_view(scalex=scalex, scaley=scaley)
        return ax

    def plotdz(self, z, distance=None, ax=None, dunit: str = 'm', doffset: float = 0,
               scalex=True, scaley=True, **kwargs):
        """Plot a node attribute list against upstream distance

        Parameters
        ----------
        z: GridObject, np.ndarray
          The node attribute list that will be plotted on the y axis

        distance: GridObject, np.ndarray
          The node attribute list that will be plotted on the x axis

        ax: matplotlib.axes.Axes, optional
            The axes in which to plot the StreamObject. If no axes are
            given, the current axes are used.

        dunit: str, optional
            The unit to plot the upstream distance. Should be either
            'm' for meters or 'km' for kilometers.

        doffset: float, optional
            An offset to be applied to the upstream distance.
            `doffset` should be in the units specified by `dunit`.

        scalex: bool, optional
            Autoscale the x-axis limits. Defaults to `True`. If the
            x-axis limits have been set manually with `set_xlim`, the
            autoscaling will not be applied regardless of the value of
            `scalex`.

        scaley: bool, optional
            Autoscale the y-axis limits. Defaults to `True`. If the
            y-axis limits have been set manually with `set_ylim`, the
            autoscaling will not be applied regardless of the value of
            `scaley`.

        **kwargs
            Additional keyword arguments are forwarded to LineCollection

        Returns
        -------
        matplotlib.axes.Axes
            The axes into which the plot as been added

        Raises
        ------
        ValueError
            If `dunit` is not one of 'm' or 'km'.

        """

        if ax is None:
            ax = plt.gca()
        z = self.ezgetnal(z)

        if distance is None:
            distance = self.upstream_distance()

        dist = self.ezgetnal(distance)

        if dunit == 'km':
            dist /= 1000
        elif dunit != 'm':
            raise ValueError("dunit must be one of 'm' or 'km'")

        dist += doffset

        collection = LineCollection(self.xy((dist, z)), **kwargs)
        ax.add_collection(collection)
        ax.autoscale_view(scalex=scalex, scaley=scaley)
        return ax

    def chitransform(self,
                     upstream_area: GridObject | np.ndarray,
                     a0: float = 1e6,
                     mn: float = 0.45,
                     k: GridObject | np.ndarray | None = None,
                     correctcellsize: bool = True):
        """Coordinate transformation using the integral approach

        Transforms the horizontal spatial coordinates of a river
        longitudinal profile using an integration in upstream
        direction of drainage area (chi, see Perron and Royden, 2013).

        Parameters
        ----------
        upstream_area : GridObject | np.ndarray
            Raster with the upstream areas. Must be the same size and
            projection as the GridObject used to create the
            StreamObject.
        a0 : float, optional
            Reference area in the same units as the upstream_area
            raster. Defaults to 100_000.
        mn : float, optional
            mn-ratio. Defaults to 0.45.
        k  : GridObject | np.ndarray | None, optional
            Erosional efficiency, which may vary spatially. If `k` is
            supplied, then `chitransform` returns the time needed for
            a signal (knickpoint) propagating upstream from the outlet
            of the stream network. If `k` has units of m^(1 - 2m) / y,
            then time will have units of y. Note that calculating the
            response time requires the assumption that n=1. Defaults
            to None, which does not use the erosional efficiency.
        correctcellsize : bool, optional
            If true, multiplies the `upstream_area` raster by
            `self.cellsize**2`. Use if `a0` has the same units of
            `self.cellsize**2` and `upstream_area` has units of
            pixels, such as the default output from
            `flow_accumulation`. If the units of `upstream_area` are
            already m^2, then set correctcellsize to False. Defaults
            to True.

        Raises
        ------
        ValueError
            If `upstream_area` or `k` does not have the right shape to
            be indexed by the `StreamObject`.
        TypeError
            If `upstream_area` or `k` does not represent a type of data that can be
            extracted into a node attribute list.
        TypeError
            If the modified upstream area is not a supported floating point type.
        """

        # Retrieve node attribute lists
        a = self.ezgetnal(upstream_area)

        if correctcellsize:
            a = a * self.cellsize**2

        # Set up k
        if k is not None:
            node_k = self.ezgetnal(k)
            a = (1 / node_k) * (1 / a)**mn
        else:
            a = (a0 / a)**mn

        # Cumulative trapezoidal integration
        weight = self.node_to_node_distance()
        c = np.zeros_like(a)
        if a.dtype == np.float32:
            _stream.streamquad_trapz_f32(c,
                                         a,
                                         self.source,
                                         self.target,
                                         weight)
        elif a.dtype == np.float64:
            _stream.streamquad_trapz_f64(c,
                                         a,
                                         self.source,
                                         self.target,
                                         weight)
        else:
            # This is probably unreachable
            raise TypeError("modified area is not a floating point object")

        return c

    def trunk(self, downstream_distance: np.ndarray | None = None,
              flow_accumulation: GridObject | None = None) -> 'StreamObject':
        """Reduces a stream network to the longest streams in each stream
        network tree (e.g. connected component). The algorithm identifies
        the main trunk by sequently tracing the maximum downstream
        distance in upstream direction.

        Parameters
        ----------
        flow_accumulation : Gridobject, optional
            A GridObject filled with flow accumulation values (as returned by
            the function FlowObject.flow_accumulation). Defaults to None.
        downstream_distance : np.ndarray, optional
            A numpy ndarray node-attribute list as generated by ezgetnal().
            This argument overwrites the flow_accumulation if used.
            Defaults to None.

        Returns
        -------
        StreamObject
            StreamObject with truncated streams.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('perfectworld')
           >>> fd = topotoolbox.FlowObject(dem)
           >>> s = topotoolbox.StreamObject(fd,threshold=1000,units='pixels')
           >>> s2 = s.klargestconncomps(1)
           >>> st = s2.trunk()
           >>> fig,ax = plt.subplots()
           >>> _= dem.plot(ax=ax,cmap="terrain")
           >>> _= s.plot(ax=ax, color='r')
           >>> _= s2.plot(ax=ax,color='k')
           >>> _= st.plot(ax=ax, color='b')
           >>> plt.show()
        """

        stream_network_size = len(self.stream)

        if not downstream_distance is None:
            pass
        elif not flow_accumulation is None:
            downstream_distance = self.ezgetnal(flow_accumulation)
        else:
            downstream_distance = self.downstream_distance()

        sparse_distance = csr_matrix(
            (downstream_distance[self.source] + 1, (self.source, self.target)),
            shape=(stream_network_size, stream_network_size))

        # Identify outlet reaches
        any_column = np.array(sparse_distance.sum(axis=0) > 0).flatten()
        any_row = np.array(sparse_distance.sum(axis=1) > 0).flatten()
        outlets = any_column & ~any_row

        trunks_max = np.argmax(sparse_distance, axis=0)

        max_neighbor = np.zeros(stream_network_size, dtype=bool)
        max_neighbor[trunks_max] = True

        trunks = np.zeros(stream_network_size, dtype=bool)
        trunks[outlets] = True

        for r in range(len(self.source) - 1, -1, -1):
            trunks[self.source[r]] = trunks[self.target[r]
                                            ] and max_neighbor[self.source[r]]

        result = self.subgraph(trunks)
        return result

    def conncomps(self) -> np.ndarray:
        """Label connected components in a stream network

        Returns
        -------
        np.ndarray
           A node attribute list containing the labels of each
           connected component. The labels range from 1 to the number
           of connected components in the network.
        """
        outlets = self.streampoi('outlets')

        labels = np.zeros(self.stream.size, dtype=np.int64)
        labels[outlets] = np.arange(np.count_nonzero(outlets)) + 1

        _stream.propagatevaluesupstream_i64(labels, self.source, self.target)

        return labels

    def klargestconncomps(self, k=1) -> 'StreamObject':
        """Extract the k largest connected components of the stream network

        Components are ordered by the number of stream network pixels.

        Parameters
        ----------
        k : integer, optional
            The number of components to keep. The default is 1

        Returns
        -------
        StreamObject
            A new StreamObject containing only the k largest connected components

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('perfectworld')
           >>> fd = topotoolbox.FlowObject(dem)
           >>> s = topotoolbox.StreamObject(fd,threshold=1000,units='pixels')
           >>> s2 = s.klargestconncomps(1)
           >>> fig, ax = plt.subplots()
           >>> _= dem.plot(ax=ax,cmap="terrain")
           >>> _= s2.plot(ax=ax,color='k')
           >>> plt.show()
        """
        nv = self.stream.size
        ne = self.source.size

        outlets = self.streampoi('outlets')

        # Count the nodes in each connected component of the stream
        # network.
        # This might be slightly inconsistent with the MATLAB
        # implementation which sorts by the number of edges in each
        # connected component.
        acc = np.ones(nv, dtype=np.float32)
        weights = np.ones(ne, dtype=np.float32)
        _stream.traverse_down_f32_add_mul(
            acc, weights, self.source, self.target)

        # Indices of the outlets in a node attribute list
        outlet_indices = np.flatnonzero(outlets)

        # Indices of the sorted accumulation values, from lowest to highest
        ixs = np.argsort(acc[outlets])

        # conncomps will be 1 for all pixels in the k largest
        # connected components
        conncomps = np.zeros(nv, dtype=np.uint8)

        # Initialize to 1 at the outlets of the k largest components
        conncomps[outlet_indices[ixs[-k:]]] = 1

        # And propagate those values from the outlets upstream
        _stream.propagatevaluesupstream_u8(conncomps, self.source, self.target)

        # Convert to boolean array so we can index with it
        conncomps_mask = conncomps > 0

        result = self.subgraph(conncomps_mask)
        return result

    def subgraph(self, nal):
        """Extract a subset of the stream network

        Parameters
        ----------
        nal: GridObject or np.ndarray
            A logical node attribute list indicating the desired
            nodes of the new stream network

        Returns
        -------
        StreamObject
            A StreamObject representing the desired subset of the
            stream network.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import numpy as np
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('perfectworld')
           >>> dem = topotoolbox.load_dem('perfectworld')
           >>> fd = topotoolbox.FlowObject(dem)
           >>> s = topotoolbox.StreamObject(fd,threshold=1000,units='pixels')
           >>> shape = dem.shape
           >>> arr = (np.arange(np.prod(shape))<np.prod(shape)//4).reshape(shape)
           >>> s2 = s.subgraph(arr)
           >>> fig,ax = plt.subplots()
           >>> _= dem.plot(ax=ax,cmap="terrain")
           >>> _= s2.plot(ax=ax,color='k')
           >>> plt.show()
        """

        nal = self.ezgetnal(nal)
        nal = nal > 0
        result = copy.copy(self)

        result.stream = self.stream[nal]

        new_indices = np.cumsum(nal) - 1

        valid_edges = nal[self.source] & nal[self.target]

        result.source = self.source[valid_edges]
        result.target = self.target[valid_edges]

        result.source = new_indices[result.source]
        result.target = new_indices[result.target]

        # MATLAB cleans the result, but this leads to a circular
        # dependency between `subgraph` and `clean` that confuses
        # things.

        # TODO(wkearn): return indices into the original node
        # attribute list
        return result

    def clean(self) -> 'StreamObject':
        """Remove disconnected nodes in stream networks

        Returns
        -------
        StreamObject
            A stream network where all isolated nodes have been removed

        Example
        -------
        >>> import topotoolbox
        >>> dem = topotoolbox.load_dem('bigtujunga')
        >>> fd = topotoolbox.FlowObject(dem)
        >>> s = topotoolbox.StreamObject(fd,threshold=1000)
        >>> sc = s.clean()
        >>> assert sc.stream.shape <= s.stream.shape
        """

        indegree = np.zeros(self.stream.size, dtype=np.uint8)
        outdegree = np.zeros(self.stream.size, dtype=np.uint8)
        _stream.edgelist_degree(indegree, outdegree, self.source, self.target)
        nal = (indegree != 0) | (outdegree != 0)

        return self.subgraph(nal)

    def upstreamto(self, nodes) -> 'StreamObject':
        """Extract the portion of the stream network upstream of the given nodes

        Parameters
        ----------
        nodes: GridObject or np.ndarray
            A logical node attribute list or grid that is True for the desired nodes.

        Returns
        -------
        StreamObject
            A stream network containing those nodes of the original
            one that are upstream of the given nodes.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import numpy as np
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('perfectworld')
           >>> fd = topotoolbox.FlowObject(dem)
           >>> s = topotoolbox.StreamObject(fd,threshold=1000,units='pixels')
           >>> confluences = s.streampoi('confluences')
           >>> s2 = s.upstreamto(confluences)
           >>> fig,ax = plt.subplots()
           >>> _= dem.plot(ax=ax,cmap="terrain")
           >>> _= s2.plot(ax=ax,color='k')
           >>> plt.show()
        """
        nal = self.ezgetnal(nodes, dtype=np.uint32)

        edges = np.ones(self.source.size, dtype=np.uint32)
        _stream.traverse_up_u32_or_and(nal, edges, self.source, self.target)

        return self.subgraph(nal)

    def downstreamto(self, nodes) -> 'StreamObject':
        """Extract the portion of the stream network downstream of the given nodes

        Parameters
        ----------
        nodes: GridObject or np.ndarray
            A logical node attribute list or grid that is True for the desired nodes.

        Returns
        -------
        StreamObject
            A stream network containing those nodes of the original
            one that are downstream of the given nodes.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import numpy as np
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('perfectworld')
           >>> fd = topotoolbox.FlowObject(dem)
           >>> s = topotoolbox.StreamObject(fd,threshold=1000,units='pixels')
           >>> confluences = s.streampoi('confluences')
           >>> s2 = s.downstreamto(confluences)
           >>> fig,ax = plt.subplots()
           >>> _= dem.plot(ax=ax,cmap="terrain")
           >>> _= s2.plot(ax=ax,color='k')
           >>> plt.show()
        """
        nal = self.ezgetnal(nodes, dtype=np.uint32)

        edges = np.ones(self.source.size, dtype=np.uint32)
        _stream.traverse_down_u32_or_and(nal, edges, self.source, self.target)

        return self.subgraph(nal)

    def gradient(self, dem, impose=False) -> 'np.ndarray':
        """Calculates the stream slope for each node in the stream
        network S based on the associated digital elevation model DEM.

        Parameters
        ----------
        dem: GridObject or np.ndarray
            A node attribute list or grid that provides the elevation that we take the gradient of.
        impose: bool
            Minima imposition to avoid negative slopes (see imposemin)

        Returns
        -------
        s
            stream gradient as node-attribute list
        """

        # get node attribute list with elevation values
        z = self.ezgetnal(dem)

        if impose:
            z = imposemin(self, z)

        # inter-node distance
        d = self.node_to_node_distance()

        # forward case
        s = np.zeros(self.stream.size)
        s[self.source] = (z[self.source]-z[self.target])/d

        return s

    def ksn(self, dem, a, impose=False, theta=0.45) -> 'np.ndarray':
        """Returns the normalized steepness index using a default concavity
        index of 0.45.

        Parameters
        ----------
        dem: GridObject or np.ndarray
            Digital elevation model
        a: GridObject
            Flow accumulation as returned by flowacc (GridObject)
        impose: bool
            Minima imposition to avoid negative slopes (see imposemin
        theta: float
            Concavity (default 0.45)

        Returns
        -------
        k
            Normalized steepness index
        """
        z = self.ezgetnal(dem)
        a = self.ezgetnal(a)

        # minima imposition to avoid negative gradients
        if impose:
            z = imposemin(self, z)

        # calculate gradient
        g = self.gradient(z)

        # upslope area
        a = a*self.cellsize**2

        # calculate k
        k = g/(a**(-theta))

        return k

    def streamorder(self, method='strahler') -> np.ndarray:
        """Calculates stream order from the StreamObject using the Strahler
        or Shreve method

        Parameters
        ----------
        s: StreamObject
        method: string
            Strahler (default) or Shreve method

        Returns
        -------
        s_order
            Stream order for each node in StreamObject
        """

        s_order = self.streampoi('channelheads')
        s_order = s_order.astype(dtype=np.float32)
        w = np.ones(self.source.shape, dtype=np.float32)

        if method.lower() == 'strahler':
            _stream.traverse_down_f32_strahler(
                s_order, w, self.source, self.target)
        elif method.lower() == 'shreve':
            _stream.traverse_down_f32_add_mul(
                s_order, w, self.source, self.target)
        else:
            raise ValueError("Invalid type. Choose 'strahler' or 'shreve'.")

        return s_order

    def crslin(self, dem, k = 1, mingradient=0.0, attachheads=False, attachtomin=False):
        """Elevation values along stream networks are frequently affected by
        large scatter, often as a result of data artifacts or errors. This
        function returns a node attribute list of elevations calculated by
        regularized smoothing.

        The algorithm written in this function follows Appendix A2 in the Schwanghart
        and Scherler 2017 paper.

        Parameters:
        -----------
        s: StreamObject

        dem: GridObject | np.ndarray
            DEM

        k: float
            positive scalar that dictates the degree of stiffness. Default is 1

        mingradient: double
            Minimum downward gradient.
            Choose carefully, because length profile may dip to steeply.
            Set this parameter to nan if you do not wish to have a monotonous
            dowstream elevation decrease.

        attachtomin: bool
            Smoothed elevations will not exceed local minima along the
            downstream path. (only applicable if 'mingradient' is not nan)

        attachheads: bool
            If true, elevations of channelheads are fixed. (only applicable
            if 'mingradient' is not nan). Note that for large K, setting
            attachheads to true can result in excessive smoothing and
            underestimation of elevation values directly downstream to channelheads.

        Returns
        -------
        zs:
            node attribute list with smoothed elevation values
        """

        # get node attribute list with elevation values
        z = self.ezgetnal(dem, dtype='double')  # elevation values of the dem
        nr = z.size
        ne = self.source.size # num of edges

        if any(np.isnan(z)):
            raise ValueError('DEM or z may not contain any NaNs')

        # CRS linear algorithm

        # Clarabel only accepts csc arrays
        # Identity Matrix
        identity_matrix = sp.eye_array(nr).tocsc()

        #########################################
        # Compute second derivative matrix (C in equation A5)
        # find upstream and downstream nodes
        ix = self.source  # upstream
        ixc = self.target  # downstraeam

        # boolean array to store nodes that are both sources and targets
        i = np.isin(ixc, ix)
        # creates a dictionary to store ix
        dic = dict(zip(ix, np.arange(self.source.shape[0])))
        # values as key and their indicies as values
        keys = np.array([dic[j] for j in ixc[i]])

        # [i-1 (downstream), i, i+1 (upstream)]
        colix = np.array([ixc[keys], ixc[i], ix[i]]).T
        nrrows = colix.shape[0]
        rowix = np.tile(np.arange(nrrows).reshape(-1, 1), 3)

        # compute distance values between nodes
        d = self.upstream_distance()

        xj = d[colix[:, 0]]  # downstream node of i
        xi = d[colix[:, 1]]
        xk = d[colix[:, 2]]  # upstream node of i

        # Dense C matrix (equation A4). Must be converted to kvxopt spare matrix for qp
        values = np.array(
            [2/((xi-xj)*(xk-xj)), -2/((xk-xi)*(xi-xj)), 2/((xk-xi)*(xk-xj))])

        # Second derivative matrix (C)
        c = sp.csc_array((values.T.flatten(), (rowix.flatten(), colix.flatten())), (nrrows, nr))

        #########################################
        # Compute s parameter (equation A7)
        delta_x = self.cellsize  # spatial resolution
        n = nr  # number of data points
        p = nrrows  # num. of second derivative equations = num. of nodes where second derivative
        # can be computed (number of i nodes --> with both upstream and dowsntream neighbours)
        s_parameter = ((delta_x)**2)*k*math.sqrt(n/p)

        #########################################
        # Compute parameters for quad programming
        # Compute matrix A and vector b (equation A9)
        a_2 = s_parameter*c  # C*s
        a_matrix = sp.vstack([identity_matrix, a_2], format = 'csc')
        b = np.concatenate([z, np.zeros(nrrows)])

        # Compute function parameters for quadratic programming (equation A11)
        f = -2*(a_matrix.T @ b)
        h_matrix = 2*(a_matrix.T @ a_matrix).tocsc()

        #########################################
        # Constraints
        # Equivalent constraint. (second constr. in equation A11)
        if attachheads:
            channelheads = self.streampoi('channelheads')
            nc = np.count_nonzero(channelheads)  # number of channelheads
            # indices of channelheads positions in array
            ids = np.where(channelheads)

            # matrix with the channelheads on main diagonal
            i_eq = sp.csc_array((channelheads[channelheads].astype(float),
                                  (np.arange(nc), ids[0])), shape=(nc, nr))
            z_eq = z[channelheads]

        else:
            i_eq = sp.csc_array((0, nr))
            z_eq = np.zeros(0)

        # Inequality constraints
        # Gradient constraint
        delta = np.array(1/(d[self.source]-d[self.target]))  # cellsize

        gradient = sp.csc_array(
            (delta, (np.arange(ne), self.target)),shape=(ne, nr)) - sp.csc_array(
                (delta, (np.arange(ne), self.source)), shape=(ne, nr))

        g_min = np.full(ne, mingradient)

        # Set up matrix G and vector g_min in equation A11. Here they are defined as M and h
        # M and h contain all inequality constraints, including upperbound (attachtomin).
        if attachtomin:
            # matrix with inequality constraints [gradient, upper bound]
            m_matrix = sp.vstack([gradient, identity_matrix], format = 'csc')
            h = np.concatenate([-g_min, z])
        else:
            m_matrix = gradient
            h = -g_min

        m_matrix_2 = sp.vstack([i_eq, m_matrix]).tocsc()
        h = np.concatenate([z_eq, h])

        # Solve quadratic programming with the constraints
        cones = [ZeroConeT(i_eq.shape[0]), NonnegativeConeT(m_matrix.shape[0])]

        settings = DefaultSettings()
        settings.verbose = False

        solver = DefaultSolver(h_matrix, f, m_matrix_2, h, cones, settings)
        sol = solver.solve()

        zs = np.array(sol.x[0:nr])

        return zs

    def quantcarve(self, dem, tau = 0.5, mingradient = 0.0, fixedoutlet = False) -> np.ndarray:
        """
        Elevation values along stream networks are frequently affected by
        large scatter, often as a result of data artifacts or errors. This
        function returns a node attribute list of elevations calculated by
        carving the DEM. Conversely to conventional carving, quantcarve will
        not run along minimas of the DEM. Instead, quantcarve returns a
        profile that runs along the tau's quantile of elevation conditional
        horizontal distance of the river profile.

        The algorithm written in this function follows Appendix A3 in the Schwanghart
        and Scherler 2017 paper.

        Parameters
        ----------
        s: StreamObject
        dem: GridObject | np.ndarray
            Digital Elevation Model
        tau: float
            Quantile. Default is 0.5
        mingradient: positive scalar
            Minimum downward gradient.
        fixedoutlet: bool
            If true, elevations of outlets are fixed.
        mingradient: float
            Minimum downward gradient.

        Returns
        -------
        zs
            Node attribute list with with smoothed elevation values
        """

        # get node attribute list with elevation values
        z = self.ezgetnal(dem, dtype = 'double') # elevation values of the dem
        nr = z.size
        ne = self.source.size # number of edges

        # Quantcarve algorithm (equation A12)
        d = self.upstream_distance() # had to use downstream distance
        f = np.vstack((tau*np.ones((nr, 1)), (1-tau)*np.ones((nr,1)), np.zeros((nr,1))))
        f = f.flatten()

        identity_matrix = sp.eye(nr).tocsc()
        zero_matrix = sp.csc_array((nr, nr))
        zero_edge_matrix = sp.csc_array((ne, nr))

        #########################################
        # Constraints
        # gradient
        delta = np.array(1/(d[self.source]-d[self.target])) # cellsize
        gradient = gradient = sp.csc_array(
            (delta, (np.arange(ne), self.target)), shape=(ne, nr)) - sp.csc_array(
                (delta, (np.arange(ne), self.source)), shape=(ne, nr))

        h_matrix = sp.csc_array((3*nr,3*nr)).tocsc() # set quadratic part of the problem
                                                     # to 0 to make it linear

        b = mingradient * np.ones(ne) # min gradient

        # bounds constraints
        lb = np.concatenate([np.zeros(nr), np.zeros(nr)])

        # stacked inequality constraints
        m_matrix = sp.block_array(
                    [[zero_edge_matrix, zero_edge_matrix, gradient],
                    [-identity_matrix, zero_matrix, zero_matrix],
                    [zero_matrix, -identity_matrix, zero_matrix]
                    ])
        h = np.concatenate([-b, lb])

        # equality constraint
        if fixedoutlet:
            outlet = self.streampoi('outlets')
            p = sp.diags_array((~outlet).astype(int), offsets = 0, shape = (nr,nr), format='csc')
            a_eq = sp.block_array([[p, -p, identity_matrix]]).tocsc()
        else:
            a_eq = sp.block_array([[identity_matrix, -identity_matrix, identity_matrix]]).tocsc()

        b_eq = np.array(z, copy = True)

        # All constraints stacked
        m_matrix_2 = sp.vstack([a_eq, m_matrix], format = 'csc')
        h_2 = np.concatenate([b_eq, h])

        #########################################
        # solve linear programming problem
        cones = [ZeroConeT(nr), NonnegativeConeT(ne + 2 * nr)]

        settings = DefaultSettings()
        settings.verbose = True

        solver = DefaultSolver(h_matrix, f, m_matrix_2, h_2, cones, settings)
        sol = solver.solve()

        #us = np.array(sol.x[0:nr])
        #vs = np.array(sol.x[nr:2*nr])
        zs = np.array(sol.x[2*nr:3*nr])

        return zs

    def crs(self, dem, tau = 0.5, k = 1, mingradient = 0.0, fixedoutlet = False) -> np.ndarray:
        """Elevation values along stream networks are frequently affected by
        large scatter, often as a result of data artifacts or errors. This
        function returns a node attribute list of smoothed elevation values
        calculated by nonparametric quantile regression with monotonicity
        constraints.

        The algorithm written in this function follows Appendix A4 in the Schwanghart
        and Scherler 2017 paper.

        Parameters:
        -----------
        s: StreamObject

        dem: GridObject | np.ndarray
            DEM

        tau: float
            Quantile. Default is 0.5

        k: float
            positive scalar that dictates the degree of stiffness. Default is 1

        mingradient: double
            Minimum downward gradient.
            Choose carefully, because length profile may dip to steeply.
            Set this parameter to nan if you do not wish to have a monotonous
            dowstream elevation decrease.

        fixedoutlet: bool
            If true, elevations of outlets are fixed.

        Returns
        ----------
        zs:
            node attribute list with smoothed elevation values
        """

        # get node attribute list with elevation values
        z = self.ezgetnal(dem, dtype=np.float64)  # node_data[:,0]
        if any(np.isnan(z)):
            raise ValueError('DEM or z may not contain any NaNs')

        nr = z.size # num. of nodes
        ne = self.source.size # num of edges

        # a special case. The stream network needs at least three nodes in a row. If
        # there are less, second derivative matrix is empty, and crs uses quantcarve instead
        if nr < 3:
            print('nr < 3')
            return self.quantcarve(dem, tau, mingradient, fixedoutlet='True')

        # CRS Algorithm
        # Identity Matrix & zero matrices
        identity_matrix = sp.eye_array(nr)
        zero_matrix = sp.csc_array((nr, nr))
        zero_edge_matrix = sp.csc_array((ne, nr))

        #########################################
        # Compute Second Derivative Matrix (C in equation A5)
        # find upstream and downstream nodes
        ix = self.source
        ixc = self.target

        # boolean array to store nodes that are both sources and targets
        i = np.isin(ixc, ix)
        dic = dict(zip(ix, np.arange(self.source.shape[0]))) # dictionary to store ix
        keys = np.array([dic[j] for j in ixc[i]]) # values as key and their indicies as values

        # [i-1 (downstream), i, i+1 (upstream)]
        colix = np.array([ixc[keys], ixc[i], ix[i]]).T
        nrrows = colix.shape[0]
        rowix = np.tile(np.arange(nrrows).reshape(-1, 1), 3)

        d = self.upstream_distance()

        xj = d[colix[:, 0]] # downstream node of i
        xi = d[colix[:, 1]]
        xk = d[colix[:, 2]] # upstream node of i

        # C matrix (equation A4)
        values = np.array([2/((xi - xj) * (xk - xj)), -2 /
                        ((xk-xi)*(xi-xj)), 2/((xk-xi)*(xk - xj))]).T
        c = sp.csc_array(
            (values.flatten(), (rowix.flatten(), colix.flatten())), (nrrows, nr))

        #########################################
        # Compute s parameter (equation A7)
        delta_x = self.cellsize
        n = nr
        p = nrrows
        s_parameter = ((delta_x)**2) * k * np.sqrt(n/p)

        #########################################
        # Compute function parameters for quadratic part of the problem
        # Compute matrix D (equation A15)
        d_2 = s_parameter*c
        d_matrix = sp.block_array([[zero_matrix], [d_2]])
        b_matrix = 2*(d_matrix.T @ d_matrix)
        h_matrix = sp.block_diag(
                [zero_matrix, zero_matrix, b_matrix], format='csc')

        # Compute function parameters for linear part of the problem
        f = np.concatenate((tau*np.ones(nr), (1-tau)*np.ones(nr), np.zeros(nr)))

        #########################################
        # Constraints
        # Gradient
        delta = 1/(d[self.source] - d[self.target])

        gradient = sp.csc_array((delta, (np.arange(ne), self.target)), shape=(
            ne, nr)) - sp.csc_array((delta, (np.arange(ne), self.source)), shape=(ne, nr))

        b = mingradient * np.ones(ne) # min gradient

        # Bounds
        lb = np.concatenate([np.zeros(nr), np.zeros(nr)])

        # Inequality Constraints
        m_matrix = sp.block_array(
            [[zero_edge_matrix, zero_edge_matrix, gradient],
            [-identity_matrix, zero_matrix, zero_matrix],
            [zero_matrix, -identity_matrix, zero_matrix]
            ])

        h = np.concatenate([-b, lb])

        # Equality Constraint
        if fixedoutlet:
            outlet = self.streampoi('outlets')
            p = sp.diags_array((~outlet).astype(np.int64), offsets = 0,
                               shape = (nr,nr), format='csc')
            a_eq = sp.block_array([[p, -p, identity_matrix]])
        else:
            a_eq = sp.block_array(
                [[identity_matrix, -identity_matrix, identity_matrix]])

        b_eq = np.array(z, copy=True)

        a_matrix = sp.vstack([a_eq, m_matrix], format='csc')
        g = np.concatenate([b_eq, h])

        cones = [ZeroConeT(nr), NonnegativeConeT(ne + 2 * nr)]

        settings = DefaultSettings()
        settings.verbose = False

        #########################################
        # Compute algorithm solution
        solver = DefaultSolver(h_matrix, f, a_matrix, g, cones, settings)
        sol = solver.solve()

        #us = np.array(sol.x[0:nr])
        #vs = np.array(sol.x[nr:2*nr])
        zs = np.array(sol.x[2*nr:3*nr])

        return zs

    def lowerenv(self, dem: GridObject | np.ndarray, kn: np.ndarray) -> np.ndarray:
        """Compute the lower convex envelope of a stream profile

        Parameters
        ----------
        dem: GridObject or np.ndarray

            The elevation data of the stream network provided either
            as a GridObject compatible with this StreamObject or as a
            node attribute list. The elevation data should be
            decreasing downstream. Preprocess with imposemin or
            quantcarve if necessary.

        kn: np.ndarray
            A logical node attribute list that is true for nodes that
            should be considered knickpoints, where the stream profile
            need not be convex.

        Returns
        -------
        np.ndarray
            A node attribute list containing the elevation of the
            smoothed profile.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import numpy as np
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('bigtujunga')
           >>> fd = topotoolbox.FlowObject(dem)
           >>> s = topotoolbox.StreamObject(fd)
           >>> s = s.klargestconncomps(1)
           >>> z = topotoolbox.imposemin(s, dem)
           >>> kn = np.zeros(len(z), dtype=bool)
           >>> ze = s.lowerenv(z, kn)
           >>> fig,ax = plt.subplots()
           >>> _= s.plotdz(dem, ax=ax, color='gray')
           >>> _= s.plotdz(ze, ax=ax, color='black')
           >>> ax.autoscale_view()
           >>> plt.show()
        """
        z = self.ezgetnal(dem, dtype=np.float32)
        d = self.upstream_distance()

        nv = len(z)
        ix = np.zeros(nv, dtype=np.int64)
        onenvelope = np.zeros(nv, dtype=np.uint8)

        _stream.lowerenv(z, kn.astype(np.uint8),
                         d, ix, onenvelope,
                         self.source, self.target)

        return z

    def knickpointfinder(self, dem: GridObject | np.ndarray,
                         knickpoints: np.ndarray | None = None,
                         tolerance: float = 100.0,
                         iterations: int | None = None) -> np.ndarray:
        """Find knickpoints in river profiles

        Rivers that adjust to changing base levels or have diverse
        lithologies often feature knickzones, i.e. pronounced convex
        sections that separate the otherwise concave equilibrium
        profile. The profile should be monotoneously decreasing (see
        imposemin, quantcarve, crs).

        This function extracts knickpoints, i.e. sharp convex sections
        in the river profile. This is accomplished by an algorithm
        that adjusts a strictly concave upward profile to an actual
        profile in the DEM or node-attribute list. The algorithm
        iteratively relaxes the concavity constraint at those nodes in
        the river profile that have the largest elevation offsets
        between the strictly concave and actual profile until the
        offset falls below a user-defined tolerance.

        Parameters
        ----------
        dem: GridObject or np.ndarray

            The elevation data of the stream network provided either
            as a GridObject compatible with this StreamObject or as a
            node attribute list. The elevation data is automatically
            processed using imposemin to ensure that it is decreasing
            downstream. If additional smoothing of the stream profile
            is needed, use a function like quantcarve or crslin and
            pass the resulting node attribute list to
            `knickpointfinder`.

        knickpoints: np.ndarray

            A logical node attribute list that is True for any stream
            network nodes that should be considered knickpoints a priori

        tolerance: float

            The maximum difference between the DEM and a modeled
            convex profile that will be considered a
            knickpoint

            Setting this higher will identify fewer knickpoints while
            setting it lower finds more knickpoints. It is set by
            default to 100 m, but users should choose a tolerance
            based on their data and needs.

        iterations: int

            The maximum number of iterations that will be
            run

            `knickpointfinder` will identify more than one knickpoint
            per iteration, so this argument does not completely
            control the number of knickpoints identified. The default
            is the number of nodes in the stream network, and the
            iteration count is always limited to the number of nodes
            in the stream network.

        Returns
        -------
        np.ndarray
            A logical node attribute list that is True for stream
            network nodes that are identified as knickpoints

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('bigtujunga')
           >>> fd = topotoolbox.FlowObject(dem)
           >>> s = topotoolbox.StreamObject(fd)
           >>> s = s.klargestconncomps(1)
           >>> z = topotoolbox.imposemin(s, dem)
           >>> kp = s.knickpointfinder(z, tolerance=50.0)
           >>> d = s.upstream_distance()
           >>> fig,ax = plt.subplots()
           >>> _= s.plotdz(dem, ax=ax, color='k')
           >>> _= ax.scatter(d[kp], z[kp])
           >>> _= ax.autoscale_view()
           >>> plt.show()
        """
        z = self.ezgetnal(dem, dtype=np.float32)
        z = imposemin(self, z)

        # The number of knickpoints and thus the number of iterations
        # is bounded by the number of nodes in the stream network.
        if iterations is None:
            iterations = self.stream.size
        iterations = min(iterations, self.stream.size)

        nv = self.stream.size

        if knickpoints is None:
            kp = np.zeros(nv, dtype=np.bool)
        else:
            kp = np.array(knickpoints, copy=True)

        for _ in np.arange(iterations):
            zs = self.lowerenv(z, kp)
            dz = z - zs
            imax = np.arange(nv)

            # These loops compute the maximum difference between the
            # observed elevation and the lower convex envelope over
            # reaches separated by knickpoints. The first loop
            # propagates the maximum value and index downstream,
            # skipping edges that terminate on a knickpoint.
            for (u, v) in zip(self.source, self.target):
                if dz[v] < dz[u] and not kp[v]:
                    dz[v] = dz[u]
                    imax[v] = imax[u]

            # The second loop propagates the maximum back upstream, so
            # each connected component is labeled with the maximum
            # difference and the node index at which that maximum
            # occurs.
            for (u, v) in zip(reversed(self.source), reversed(self.target)):
                if dz[u] < dz[v] and not kp[v]:
                    dz[u] = dz[v]
                    imax[u] = imax[v]

            # These loops could be implemented in libtopotoolbox for
            # speed. The maximum-over-reaches logic could be useful
            # for other applications.

            # The unique node indices are the potential knickpoints
            # for each reach. We select only those whose difference
            # exceeds the user-provided tolerance.
            potential_knicks = np.unique(imax)
            new_knicks = dz[potential_knicks] >= tolerance

            # If no new knickpoints have been found, we stop
            # searching.
            if np.count_nonzero(new_knicks) == 0:
                break

            kp[potential_knicks] = new_knicks

        return kp

    # 'Magic' functions:
    # ------------------------------------------------------------------------

    def __len__(self):
        return len(self.stream)

    def __iter__(self):
        return iter(self.stream)

    def __getitem__(self, index):
        return self.stream[index]

    def __setitem__(self, index, value):
        try:
            value = np.float32(value)
        except (ValueError, TypeError):
            raise TypeError(
                f"{value} can't be converted to float32.") from None

        self.stream[index] = value

    def __array__(self):
        return self.stream

    def __str__(self):
        return str(self.stream)



import warnings
import pytest

import numpy as np
import opensimplex

import topotoolbox as topo


def issubgraph(s1: topo.StreamObject, s2: topo.StreamObject):
    """Test whether s1 represents a subgraph of s2
    """
    e1 = set(zip(zip(*s1.source_indices), zip(*s1.target_indices)))
    e2 = set(zip(zip(*s2.source_indices), zip(*s2.target_indices)))
    return e1 <= e2


def isequivalent(s1: topo.StreamObject, s2: topo.StreamObject):
    v1 = set(zip(*s1.node_indices))
    v2 = set(zip(*s2.node_indices))

    e1 = set(zip(zip(*s1.source_indices), zip(*s1.target_indices)))
    e2 = set(zip(zip(*s2.source_indices), zip(*s2.target_indices)))

    return v1 == v2 and e1 == e2


@pytest.fixture(name="wide_dem")
def fixture_wide_dem():
    yield topo.gen_random(rows=64, columns=128)


@pytest.fixture(name="wide_flow")
def fixture_wide_flow(wide_dem):
    yield topo.FlowObject(wide_dem)


@pytest.fixture(name="wide_stream")
def fixture_wide_stream(wide_flow):
    yield topo.StreamObject(wide_flow)


@pytest.fixture(name="tall_dem")
def fixture_tall_dem():
    yield topo.gen_random(rows=128, columns=64)


@pytest.fixture(name="tall_flow")
def fixture_tall_flow(tall_dem):
    yield topo.FlowObject(tall_dem)


@pytest.fixture(name="tall_stream")
def fixture_tall_stream(tall_flow):
    yield topo.StreamObject(tall_flow)


@pytest.fixture
def order_dems():
    opensimplex.seed(12)

    x = np.arange(0, 128)
    y = np.arange(0, 256)

    cdem = topo.GridObject()
    cdem.z = np.array(
        64 * (opensimplex.noise2array(x/13, y/13) + 1), dtype=np.float32)
    cdem.cellsize = 13.0

    fdem = topo.GridObject()
    fdem.z = np.asfortranarray(cdem.z)
    fdem.cellsize = 13.0

    return {"cdem": cdem, "fdem": fdem}


@pytest.fixture(name="cfd")
def fixture_cfd(order_dems):
    cdem = order_dems["cdem"]
    yield topo.FlowObject(cdem)


@pytest.fixture(name="ffd")
def fixture_ffd(order_dems):
    fdem = order_dems["fdem"]
    yield topo.FlowObject(fdem)


@pytest.fixture(name="cs")
def fixture_cs(cfd):
    yield topo.StreamObject(cfd)


@pytest.fixture(name="fs")
def fixture_fs(ffd):
    yield topo.StreamObject(ffd)


def test_constructors():
    grid_obj = topo.gen_random(rows=64, columns=64)
    flow_obj = topo.FlowObject(grid_obj)

    # The first positional argument has to be a FlowObject.
    with pytest.raises(TypeError):
        topo.StreamObject(1)

    # Units can only be 'pixels','km2','m2' or 'mapunits'
    with pytest.raises(ValueError):
        topo.StreamObject(flow_obj, units=' ')

    # The threshold ndarray should have the same shape as flow_obj.
    with pytest.raises(ValueError):
        topo.StreamObject(flow_obj, threshold=np.zeros(shape=(1, 1)))

    # The threshold GridObject should have the same shape as flow_obj.
    with pytest.raises(ValueError):
        topo.StreamObject(flow_obj, threshold=topo.gen_random(
            rows=1, columns=1))

    # An empty stream_pixels should result in all arrays in the StreamObject
    # being empty (len = 0).
    arr = np.zeros_like(grid_obj.z)
    stream_obj = topo.StreamObject(flow_obj, stream_pixels=arr)
    assert stream_obj.stream.size == 0
    assert stream_obj.source.size == 0
    assert stream_obj.target.size == 0
    assert stream_obj.direction.size == 0

    # When calling with a stream_pixels, threshold should be ignored and a
    # warning should be thrown.
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")  # Catch all warnings

        stream_obj = topo.StreamObject(
            flow_obj, stream_pixels=arr, threshold=1000)

        # Check that a warning was raised
        assert len(w) > 0
        assert issubclass(w[-1].category, Warning)
        assert "threshold will be ignored" in str(w[-1].message)


def test_streamobject_order(cs, fs):
    assert isequivalent(cs, fs)


def test_streamobject_streampixels_order(cfd, ffd):
    ca = cfd.flow_accumulation()
    fa = ffd.flow_accumulation()

    csp = ca.z > 100
    fsp = fa.z > 100

    cs = topo.StreamObject(cfd, stream_pixels=csp)
    fs = topo.StreamObject(ffd, stream_pixels=fsp)

    assert isequivalent(cs, fs)


def test_streamobject_sizes(tall_dem, tall_flow, tall_stream,
                            wide_dem, wide_flow, wide_stream):
    assert topo.validate_alignment(tall_dem, tall_flow)
    assert topo.validate_alignment(tall_dem, tall_stream)

    assert topo.validate_alignment(wide_dem, wide_flow)
    assert topo.validate_alignment(wide_dem, wide_stream)

    assert not topo.validate_alignment(wide_dem, tall_flow)
    assert not topo.validate_alignment(wide_dem, tall_stream)

    assert (wide_stream.target.size == wide_stream.source.size ==
            wide_stream.direction.size)
    assert (tall_stream.target.size == tall_stream.source.size ==
            tall_stream.direction.size)

    # Ensure that no index in stream exceeds the max possible index in the grid
    assert np.max(
        wide_stream.stream) <= wide_stream.shape[0] * wide_stream.shape[1]
    assert np.max(
        tall_stream.stream) <= tall_stream.shape[0] * tall_stream.shape[1]


def test_distance_order(cfd, ffd, cs, fs):
    cd = cs.node_to_node_distance()
    fd = fs.node_to_node_distance()

    cdg = np.zeros(cfd.shape)
    fdg = np.zeros(ffd.shape)

    cdg[cs.source_indices] = cd
    fdg[fs.source_indices] = fd

    assert np.array_equal(cdg, fdg)


def test_downstream_distance_order(cfd, ffd, cs, fs):
    cd = cs.downstream_distance()
    fd = fs.downstream_distance()

    cdg = np.zeros(cfd.shape)
    fdg = np.zeros(ffd.shape)

    cdg[cs.node_indices] = cd
    fdg[fs.node_indices] = fd

    assert np.array_equal(cdg, fdg)


def test_run_chitransform(tall_flow, wide_flow, tall_stream, wide_stream):
    tall_acc = tall_flow.flow_accumulation()
    wide_acc = wide_flow.flow_accumulation()

    tall_stream.chitransform(tall_acc)
    wide_stream.chitransform(wide_acc)


@pytest.mark.parametrize("acctype", [np.float32, np.float64])
@pytest.mark.parametrize("k", [None, 1e-5])
def test_chitransform_order(cfd, ffd, cs, fs, k, acctype):
    ca = np.asarray(cfd.flow_accumulation(), dtype=acctype)
    fa = np.asarray(ffd.flow_accumulation(), dtype=acctype)

    cchi = cs.chitransform(ca, k = k)
    fchi = fs.chitransform(fa, k = k)

    cchimap = np.zeros(cs.shape)
    fchimap = np.zeros(fs.shape)

    cchimap[cs.node_indices] = cchi
    fchimap[fs.node_indices] = fchi

    assert np.array_equal(cchimap, fchimap)


def test_stream_subgraphs(tall_stream, wide_stream):
    tall_trunk = tall_stream.trunk()
    tall_k1 = tall_stream.klargestconncomps(1)
    tall_k1_trunk = tall_k1.trunk()

    assert issubgraph(tall_trunk, tall_stream)
    assert issubgraph(tall_k1, tall_stream)
    assert issubgraph(tall_k1_trunk, tall_stream)
    assert issubgraph(tall_k1_trunk, tall_k1)
    assert not issubgraph(tall_trunk, tall_k1)

    assert topo.validate_alignment(tall_trunk, tall_stream)
    assert topo.validate_alignment(tall_k1, tall_stream)
    assert topo.validate_alignment(tall_k1_trunk, tall_stream)

    wide_trunk = wide_stream.trunk()
    wide_k1 = wide_stream.klargestconncomps(1)
    wide_k1_trunk = wide_k1.trunk()

    assert issubgraph(wide_trunk, wide_stream)
    assert issubgraph(wide_k1, wide_stream)
    assert issubgraph(wide_k1_trunk, wide_stream)
    assert issubgraph(wide_k1_trunk, wide_k1)
    assert not issubgraph(wide_trunk, wide_k1)

    assert topo.validate_alignment(wide_trunk, wide_stream)
    assert topo.validate_alignment(wide_k1, wide_stream)
    assert topo.validate_alignment(wide_k1_trunk, wide_stream)

    assert not issubgraph(wide_trunk, tall_stream)
    assert not issubgraph(tall_trunk, wide_stream)

@pytest.mark.parametrize("dist", [True, False])
@pytest.mark.parametrize("acc", [True, False])
def test_trunk_order(cs, fs, cfd, dist, acc):
    acc = cfd.flow_accumulation() if acc else None
    cdist = cs.downstream_distance() if dist else None
    fdist = fs.downstream_distance() if dist else None

    ctrunk = cs.trunk(downstream_distance=cdist, flow_accumulation=acc)
    ftrunk = fs.trunk(downstream_distance=fdist, flow_accumulation=acc)

    assert isequivalent(ctrunk, ftrunk)


def test_klargestconncomps_order(cs, fs):
    ck1 = cs.klargestconncomps()
    fk1 = fs.klargestconncomps()

    assert isequivalent(ck1, fk1)


def test_subgraph_order(order_dems, cs, fs):
    # Subgraph can take GridObjects/ndarrays as input. Ensure that
    # those work regardless of memory order
    cdem = order_dems["cdem"]
    cb = np.array(cdem) > 50
    fb = np.asfortranarray(cb)

    # Try every combination
    csc = cs.subgraph(cb)
    csf = cs.subgraph(fb)
    fsc = fs.subgraph(cb)
    fsf = fs.subgraph(fb)

    assert isequivalent(csc, csf)
    assert isequivalent(csc, fsc)
    assert isequivalent(csc, fsf)


def test_ezgetnal(tall_dem, tall_stream):
    z = tall_stream.ezgetnal(tall_dem)
    z2 = tall_stream.ezgetnal(z)
    z3 = tall_stream.ezgetnal(z, dtype=np.float64)

    # ezgetnal should be idempotent
    assert np.array_equal(z, z2)
    assert np.array_equal(z, z3)

    # ezgetnal should always return a copy
    assert z is not z2
    assert z is not z3

    assert z.dtype == tall_dem.z.dtype
    assert z2.dtype == tall_dem.z.dtype
    # ezgetnal with the dtype argument should return array of that type
    assert z3.dtype is np.dtype(np.float64)


def test_ezgetnal_order(order_dems, cs, fs):
    cdem = order_dems["cdem"]
    fdem = order_dems["fdem"]

    cz = cs.ezgetnal(cdem)
    fz = fs.ezgetnal(fdem)

    cdz = np.zeros(cs.shape)
    fdz = np.zeros(fs.shape)

    cdz[cs.node_indices] = cz
    fdz[fs.node_indices] = fz

    assert np.array_equal(cdz, fdz)


def test_streampoi_order(cs, fs):
    ch = cs.streampoi('channelheads')
    co = cs.streampoi('outlets')
    cc = cs.streampoi('confluences')
    cb = cs.streampoi('bconfluences')

    fh = fs.streampoi('channelheads')
    fo = fs.streampoi('outlets')
    fc = fs.streampoi('confluences')
    fb = fs.streampoi('bconfluences')

    cp = np.zeros(cs.shape, dtype=np.uint32)
    fp = np.zeros(fs.shape, dtype=np.uint32)

    cp[cs.node_indices_where(ch)] |= 1
    cp[cs.node_indices_where(co)] |= 2
    cp[cs.node_indices_where(cc)] |= 4
    cp[cs.node_indices_where(cb)] |= 8

    fp[fs.node_indices_where(fh)] |= 1
    fp[fs.node_indices_where(fo)] |= 2
    fp[fs.node_indices_where(fc)] |= 4
    fp[fs.node_indices_where(fb)] |= 8

    assert np.array_equal(cp, fp)

def test_confluences(cs):
    c = cs.streampoi('confluences')
    b = cs.streampoi('bconfluences')

    # Every bconfluence has a confluence as a downstream neighbor
    assert np.array_equal(b[cs.source], c[cs.target])

def test_subgraph(tall_stream, wide_stream):
    ############
    # Tall DEM #
    ############

    d = tall_stream.downstream_distance()
    nal = d > 10

    sub = tall_stream.subgraph(nal)

    # The subgraph should have no more vertices than are true in the
    # node attribute list.
    assert np.size(sub.stream) <= np.count_nonzero(nal)

    # The subgraph should have no more edges than the original network
    assert np.size(sub.source) <= np.size(tall_stream.source)

    # The nodes of the subgraph should be a subset of the nodes of the
    # original graph
    assert set(sub.stream) <= set(tall_stream.stream)

    # The edges of the subgraph should be a subset of the edges of the
    # original graph. This is what issubgraph tests.
    assert issubgraph(sub, tall_stream)

    # Subgraph should be idempotent
    nal_all = np.ones(tall_stream.stream.size, dtype=bool)
    sub_all = tall_stream.subgraph(nal_all)
    assert issubgraph(sub, tall_stream)
    assert np.array_equal(sub_all.stream, tall_stream.stream)
    assert np.array_equal(sub_all.source, tall_stream.source)
    assert np.array_equal(sub_all.target, tall_stream.target)

    # Subgraph of an empty subset should be empty
    nal_none = np.zeros(tall_stream.stream.size, dtype=bool)
    sub_none = tall_stream.subgraph(nal_none)
    assert issubgraph(sub, tall_stream)
    assert np.size(sub_none.stream) == 0
    assert np.size(sub_none.source) == 0
    assert np.size(sub_none.target) == 0

    ############
    # Wide DEM #
    ############
    d = wide_stream.downstream_distance()
    nal = d > 10

    sub = wide_stream.subgraph(nal)

    # The subgraph should have no more vertices than are true in the
    # node attribute list.
    assert np.size(sub.stream) <= np.count_nonzero(nal)

    # The subgraph should have no more edges than the original network
    assert np.size(sub.source) <= np.size(wide_stream.source)

    # The nodes of the subgraph should be a subset of the nodes of the
    # original graph
    assert set(sub.stream) <= set(wide_stream.stream)

    # The edges of the subgraph should be a subset of the edges of the
    # original graph. This is what issubgraph tests.
    assert issubgraph(sub, wide_stream)

    # Subgraph should be idempotent
    nal_all = np.ones(wide_stream.stream.size, dtype=bool)
    sub_all = wide_stream.subgraph(nal_all)
    assert issubgraph(sub, wide_stream)
    assert np.array_equal(sub_all.stream, wide_stream.stream)
    assert np.array_equal(sub_all.source, wide_stream.source)
    assert np.array_equal(sub_all.target, wide_stream.target)

    # Subgraph of an empty subset should be empty
    nal_none = np.zeros(wide_stream.stream.size, dtype=bool)
    sub_none = wide_stream.subgraph(nal_none)
    assert issubgraph(sub, wide_stream)
    assert np.size(sub_none.stream) == 0
    assert np.size(sub_none.source) == 0
    assert np.size(sub_none.target) == 0


def test_stream_channelheads(tall_flow, wide_flow, tall_stream, wide_stream):
    # Tall DEM
    assert topo.validate_alignment(tall_flow, tall_stream)

    channel_heads = tall_stream.streampoi("channelheads")

    s2 = topo.StreamObject(
        tall_flow, channelheads=tall_stream.node_indices_where(channel_heads))

    assert topo.validate_alignment(tall_flow, s2)

    assert np.array_equal(s2.stream[s2.source],
                          tall_stream.stream[tall_stream.source])
    assert np.array_equal(s2.stream[s2.target],
                          tall_stream.stream[tall_stream.target])

    # Wide DEM
    assert topo.validate_alignment(wide_flow, wide_stream)

    channel_heads = wide_stream.streampoi("channelheads")

    s2 = topo.StreamObject(
        wide_flow, channelheads=wide_stream.node_indices_where(channel_heads))

    assert topo.validate_alignment(wide_flow, s2)

    assert np.array_equal(s2.stream[s2.source],
                          wide_stream.stream[wide_stream.source])
    assert np.array_equal(s2.stream[s2.target],
                          wide_stream.stream[wide_stream.target])


def test_streamobject_ch_order(cfd, ffd, cs, fs):
    cch = cs.streampoi("channelheads")
    fch = fs.streampoi("channelheads")

    cs2 = topo.StreamObject(cfd, channelheads=cs.node_indices_where(cch))
    fs2 = topo.StreamObject(ffd, channelheads=fs.node_indices_where(fch))

    assert isequivalent(cs2, fs2)


def test_stream_downstreamto(tall_flow, tall_stream):
    ch = tall_stream.streampoi("channelheads")

    sc = topo.StreamObject(
        tall_flow, channelheads=tall_stream.node_indices_where(ch))
    sd = tall_stream.downstreamto(ch)

    # These two stream networks should be equivalent
    assert len(set(sd.stream).symmetric_difference(set(sc.stream))) == 0


def test_stream_upstreamto(tall_stream):

    # We clean here in case any 1 pixel streams exist in s.
    # They won't be identified as outlets, so the reconstructed stream
    # network would not be identical to the original.
    s = tall_stream.clean()

    outlets = s.streampoi("outlets")

    su = s.upstreamto(outlets)

    # These two stream networks should be equivalent
    assert len(set(s.stream).symmetric_difference(set(su.stream))) == 0


def test_upstreamto_order(order_dems, cs, fs):
    cdem = order_dems["cdem"]
    cs = cs.clean()
    fs = fs.clean()

    b = np.array(cdem) > 50

    cb = cs.upstreamto(b)
    fb = fs.upstreamto(b)

    # These assertions make sure that this test is doing
    # something. Otherwise we may need to change the threshold above.
    assert not isequivalent(cb, cs)
    assert not isequivalent(fb, fs)

    assert isequivalent(cb, fb)


def test_downstreamto_order(order_dems, cs, fs):
    cdem = order_dems["cdem"]
    b = np.array(cdem) > 50

    cb = cs.downstreamto(b)
    fb = fs.downstreamto(b)

    assert not isequivalent(cb, cs)
    assert not isequivalent(fb, fs)

    assert isequivalent(cb, fb)


def test_stream_imposemin(tall_dem, tall_stream, wide_dem, wide_stream):
    # Tall DEM
    original_z = tall_stream.ezgetnal(tall_dem)

    for minimum_slope in [0.0, 0.001, 0.01, 0.1]:
        minz = topo.imposemin(tall_stream, tall_dem, minimum_slope)

        # imposemin should not modify the original array
        assert np.array_equal(original_z, tall_stream.ezgetnal(tall_dem))

        # The carved dem should not be above the original
        assert np.all(minz <= tall_stream.ezgetnal(tall_dem))

        # The gradient along the flow network should be greater than or
        # equal to the defined slope within some numerical error
        g = (minz[tall_stream.source] - minz[tall_stream.target]) / \
            tall_stream.node_to_node_distance()
        assert np.all(g >= minimum_slope - 1e-6)

    # Wide DEM
    original_z = wide_stream.ezgetnal(wide_dem)

    for minimum_slope in [0.0, 0.001, 0.01, 0.1]:
        minz = topo.imposemin(wide_stream, wide_dem, minimum_slope)

        # imposemin should not modify the original array
        assert np.array_equal(original_z, wide_stream.ezgetnal(wide_dem))

        # The carved dem should not be above the original
        assert np.all(minz <= wide_stream.ezgetnal(wide_dem))

        # The gradient along the flow network should be greater than or
        # equal to the defined slope within some numerical error
        g = (minz[wide_stream.source] - minz[wide_stream.target]) / \
            wide_stream.node_to_node_distance()
        assert np.all(g >= minimum_slope - 1e-6)


def test_imposemin_order(order_dems, cs, fs):
    cdem = order_dems["cdem"]
    fdem = order_dems["fdem"]

    cminslope = topo.imposemin(cs, cdem, minimum_slope=0.001)
    cz = np.zeros_like(cdem)
    cz[cs.node_indices] = cminslope

    fminslope = topo.imposemin(fs, fdem, minimum_slope=0.001)
    fz = np.zeros_like(fdem)
    fz[fs.node_indices] = fminslope

    assert np.array_equal(cz, fz)


def test_gradient(wide_dem):
    fd = topo.FlowObject(wide_dem)
    s = topo.StreamObject(fd)
    g = s.gradient(wide_dem, impose=True)

    assert np.all(g >= 0)

@pytest.mark.parametrize("impose", [True, False])
def test_gradient_order(order_dems, cs, fs, impose):
    cdem = order_dems["cdem"]

    cg = cs.gradient(cdem, impose=impose)
    cgz = np.zeros_like(cdem)
    cgz[cs.node_indices] = cg

    fg = fs.gradient(cdem, impose=impose)
    fgz = np.zeros_like(cdem)
    fgz[fs.node_indices] = fg

    assert np.array_equal(fgz, cgz)

def test_ksn(wide_dem):
    fd = topo.FlowObject(wide_dem)
    A = fd.flow_accumulation()
    s = topo.StreamObject(fd)
    theta = 0.45
    g = s.gradient(wide_dem, impose=True)
    k = s.ksn(wide_dem, A, theta)

@pytest.mark.parametrize("impose", [True, False])
def test_ksn_order(order_dems, cfd, cs, fs, impose):
    dem = order_dems["cdem"]

    a = cfd.flow_accumulation()

    ck = cs.ksn(dem, a, impose=impose)
    cks = np.zeros_like(dem)
    cks[cs.node_indices] = ck

    fk = fs.ksn(dem, a, impose=impose)
    fks = np.zeros_like(dem)
    fks[fs.node_indices] = fk

    assert np.array_equal(cks, fks)

@pytest.mark.parametrize("method", ['strahler', 'shreve'])
def test_streamorder_order(cfd, ffd, cs, fs, method):
    cds = cs.streamorder(method=method)
    fds = fs.streamorder(method=method)

    cdg = np.zeros(cfd.shape)
    fdg = np.zeros(ffd.shape)

    cdg[cs.node_indices] = cds
    fdg[fs.node_indices] = fds

    assert np.array_equal(cdg, fdg)


def test_crslin(wide_dem, wide_stream):
    s = wide_stream.klargestconncomps()

    z = s.ezgetnal(wide_dem, dtype='double')
    zs = s.crslin(wide_dem, k=1, mingradient=0.01,
                  attachtomin=True, attachheads=True)

    # attachtomin: zs <= z
    assert np.all(zs <= z)

    # attachheads: zs(channelheads) = z(channelheads)
    channelheads = s.streampoi('channelheads')
    assert np.allclose(zs[channelheads], z[channelheads])

    # gradient: gradient > mingradient
    gradient = (zs[s.source] - zs[s.target]) / s.node_to_node_distance()
    mingradient = 0.01

    assert np.all((gradient - mingradient) >= -1e-6)


def test_crslin_noattachheads(wide_dem, wide_stream):
    z = wide_stream.ezgetnal(wide_dem, dtype='double')
    zs = wide_stream.crslin(wide_dem, k=1, mingradient=0.01,
                            attachtomin=True, attachheads=False)

    assert np.all(zs <= z)

    # gradient: gradient > mingradient
    gradient = (zs[wide_stream.source] -
                zs[wide_stream.target]) / wide_stream.node_to_node_distance()
    mingradient = 0.01
    assert np.all((gradient - mingradient) >= -1e-6)

@pytest.mark.parametrize("attachheads", [True, False])
@pytest.mark.parametrize("attachtomin", [True, False])
def test_crslin_order(order_dems, cs, fs, attachtomin, attachheads):
    cdem = order_dems["cdem"]
    fdem = order_dems["fdem"]

    cp = cs.crslin(cdem, attachtomin=attachtomin, attachheads=attachheads)
    cz = np.zeros_like(cdem)
    cz[cs.node_indices] = cp

    fp = fs.crslin(fdem, attachtomin=attachtomin, attachheads=attachheads)
    fz = np.zeros_like(fdem)
    fz[fs.node_indices] = fp

    assert np.array_equal(cz, fz)


def test_quantcarve(wide_dem, wide_stream):
    s = wide_stream.klargestconncomps()
    z = s.ezgetnal(wide_dem, dtype='double')

    zs = s.quantcarve(wide_dem, tau=0.5, mingradient=0.01, fixedoutlet=True)

    # outlets: zs[outlets] = z[outlets]
    outlets = s.streampoi('outlets')
    assert np.all(zs[outlets] == z[outlets])

    # gradient: gradient > mingradient
    gradient = (zs[s.source] - zs[s.target]) / s.node_to_node_distance()
    mingradient = 0.01
    assert np.all((gradient - mingradient) >= -1e-6)

@pytest.mark.parametrize("fixed", [True, False])
def test_quantcarve_order(order_dems, cs, fs, fixed):
    cdem = order_dems["cdem"]
    fdem = order_dems["fdem"]

    cp = cs.quantcarve(cdem, fixedoutlet=fixed)
    cz = np.zeros_like(cdem)
    cz[cs.node_indices] = cp

    fp = fs.quantcarve(fdem, fixedoutlet=fixed)
    fz = np.zeros_like(fdem)
    fz[fs.node_indices] = fp

    assert np.array_equal(cz, fz)

def test_crs(wide_dem, wide_stream):
    s = wide_stream.klargestconncomps()
    z = s.ezgetnal(wide_dem, dtype='double')

    zs = s.crs(wide_dem, tau=0.5, mingradient=0.01, fixedoutlet=True)

    # outlets: zs[outlets] = z[outlets]
    outlets = s.streampoi('outlets')
    assert np.allclose(zs[outlets], z[outlets])

    # gradient: gradient > mingradient
    gradient = (zs[s.source] - zs[s.target]) / s.node_to_node_distance()
    mingradient = 0.01
    assert np.all((gradient - mingradient) >= -1e-6)

@pytest.mark.parametrize("fixed", [True, False])
def test_crs_order(order_dems, cs, fs, fixed):
    cdem = order_dems["cdem"]
    fdem = order_dems["fdem"]

    cp = cs.crs(cdem, fixedoutlet=fixed)
    cz = np.zeros_like(cdem)
    cz[cs.node_indices] = cp

    fp = fs.crs(fdem, fixedoutlet=fixed)
    fz = np.zeros_like(fdem)
    fz[fs.node_indices] = fp

    assert np.array_equal(cz, fz)

def test_lowerenv_convex(wide_dem, wide_stream):
    s = wide_stream.klargestconncomps(1)

    z = topo.imposemin(s, wide_dem)

    kn = np.arange(len(z)) % 13 == 0

    ze = s.lowerenv(z, kn)

    g = np.zeros(len(z))
    g[s.source] = (ze[s.source] - ze[s.target]) / s.node_to_node_distance()

    # The gradient of any incoming edge to a node should be greater
    # than the gradient of its outgoing edge, with some room for
    # numerical errors, except at knickpoints
    assert np.all((g[s.source] >= g[s.target] - 1e-6) + kn[s.target])


def test_lowerenv_order(order_dems, cs, fs):
    cdem = order_dems["cdem"]
    cs = cs.klargestconncomps(1)

    cz = topo.imposemin(cs, cdem)

    ckn = np.arange(len(cz)) % 13 == 0

    cze = cs.lowerenv(cz, ckn)

    cg = np.zeros(len(cz))
    cg[cs.source] = (cze[cs.source] - cze[cs.target]) / cs.node_to_node_distance()

    # The gradient of any incoming edge to a node should be greater
    # than the gradient of its outgoing edge, with some room for
    # numerical errors, except at knickpoints
    assert np.all((cg[cs.source] >= cg[cs.target] - 1e-5) + ckn[cs.target])

    fdem = order_dems["fdem"]
    fs = fs.klargestconncomps(1)

    fz = topo.imposemin(fs, fdem)

    # Make sure the same knickpoints are used in both arrays
    kn = np.zeros(cs.shape, dtype=bool)
    kn[cs.node_indices] = ckn
    fkn = kn[fs.node_indices]

    fze = fs.lowerenv(fz, fkn)

    fg = np.zeros(len(fz))
    fg[fs.source] = (fze[fs.source] - fze[fs.target]) / fs.node_to_node_distance()

    assert np.all((fg[fs.source] >= fg[fs.target] - 1e-5) + fkn[fs.target])

    czz = np.zeros(cs.shape)
    fzz = np.zeros(fs.shape)

    czz[cs.node_indices] = cze
    fzz[fs.node_indices] = fze

    assert np.allclose(czz, fzz)


def test_conncomps(wide_stream):
    s = wide_stream.clean()
    labels = s.conncomps()

    # Each node is in the same connected component as its downstream
    # neighbor.
    assert np.array_equal(labels[s.source],
                          labels[s.target])

    # There should be at least one connected component
    assert np.max(labels) > 0
    # The minimum label should be 1
    assert np.min(labels) == 1


def test_conncomps_order(cs, fs):
    clg = np.zeros(cs.shape)
    clg[cs.node_indices] = cs.conncomps()

    flg = np.zeros(cs.shape)
    flg[fs.node_indices] = fs.conncomps()

    # Labels are identical up to a bijection. See
    # tests/test_flow_object.py (test_drainagebasins_order) for
    # discussion on this approach.
    cus, cidxs, cinvs = np.unique(clg, return_index=True, return_inverse=True)
    fus, fidxs, finvs = np.unique(flg, return_index=True, return_inverse=True)

    # We use flatten because older versions of numpy
    # return different `invs` of different dimensions.
    assert np.array_equal(flg.flatten(),
                          fus[np.take(finvs, cidxs)][cinvs].flatten())
    assert np.array_equal(clg.flatten(),
                          cus[np.take(cinvs, fidxs)][finvs].flatten())


def test_knickpoints_none(wide_dem):
    fd = topo.FlowObject(wide_dem)
    s = topo.StreamObject(fd)
    s = s.klargestconncomps(1)

    kp0 = np.arange(s.stream.size) % 13 == 0

    kp = s.knickpointfinder(wide_dem, knickpoints=kp0,
                            tolerance=0.0, iterations=0)

    assert np.array_equal(kp, kp0)


def test_knickpoints_one(wide_dem):
    fd = topo.FlowObject(wide_dem)
    s = topo.StreamObject(fd)
    s = s.klargestconncomps(1)

    kp = s.knickpointfinder(wide_dem, tolerance=0.0, iterations=1)

    assert np.count_nonzero(kp) == 1


def test_knickpoints_kpcopy(wide_dem):
    fd = topo.FlowObject(wide_dem)
    s = topo.StreamObject(fd)
    s = s.klargestconncomps(1)

    kp0 = np.arange(s.stream.size) % 13 == 0
    kp_copy = np.array(kp0, copy=True)

    kp = s.knickpointfinder(wide_dem, knickpoints=kp_copy,
                            tolerance=0.0, iterations=1)

    assert np.array_equal(kp_copy, kp0)

    assert np.count_nonzero(kp) >= np.count_nonzero(kp0)

def test_knickpoints_order(order_dems, cs, fs):
    cdem = order_dems["cdem"]

    ckp = cs.knickpointfinder(cdem, tolerance=1.0)
    cks = np.zeros_like(cdem, dtype=bool)
    cks[cs.node_indices] = ckp

    fkp = fs.knickpointfinder(cdem, tolerance=1.0)
    fks = np.zeros_like(cdem, dtype=bool)
    fks[fs.node_indices] = fkp

    assert np.array_equal(cks, fks)

def test_knickpoints_provided_order(order_dems, cs, fs):
    cdem = order_dems["cdem"]

    # This is just a way of sampling some pre-existing knickpoints
    # consistently across the two memory orders
    kp0 = np.asarray(cdem.zscore()) > 2.0

    ckp = cs.knickpointfinder(cdem, tolerance=1.0, knickpoints=cs.ezgetnal(kp0))
    cks = np.zeros_like(cdem, dtype=bool)
    cks[cs.node_indices] = ckp

    fkp = fs.knickpointfinder(cdem, tolerance=1.0, knickpoints=fs.ezgetnal(kp0))
    fks = np.zeros_like(cdem, dtype=bool)
    fks[fs.node_indices] = fkp

    assert np.array_equal(cks, fks)

def test_togeodataframe(wide_stream):
    wide_stream.to_geodataframe()

def test_gridmask_order(cs, fs):
    assert np.array_equal(cs.gridmask, fs.gridmask)

def test_distance(cs):
    # We need to clean because one pixel streams will have zero
    # max_from_ch distance but infinite min_from_ch (because they are
    # not considered channelheads).
    cs = cs.clean()

    assert np.array_equal(cs.distance('from_outlet'), cs.upstream_distance())
    assert np.array_equal(cs.distance('max_from_ch'), cs.downstream_distance())

    assert np.all(cs.distance('min_from_ch') <= cs.distance('max_from_ch'))

    d = cs.distance('node_to_node')
    assert np.array_equal(d[cs.source], cs.node_to_node_distance())



import copy
import pytest

import numpy as np
import opensimplex

import topotoolbox as topo


def isequivalent(s1 : topo.FlowObject, s2 : topo.FlowObject):
    e1 = set(zip(zip(*s1.source_indices),zip(*s1.target_indices)))
    e2 = set(zip(zip(*s2.source_indices),zip(*s2.target_indices)))

    return (s1.shape == s2.shape) and (e1 == e2)


@pytest.fixture
def wide_dem():
    return topo.gen_random(rows=64, columns=128, seed=12)


@pytest.fixture
def order_dems():
    opensimplex.seed(12)

    x = np.arange(0, 128)
    y = np.arange(0, 256)

    cdem = topo.GridObject()
    cdem.z = np.array(
        64 * (opensimplex.noise2array(x/13, y/13) + 1), dtype=np.float32)
    cdem.cellsize = 13.0

    fdem = topo.GridObject()
    fdem.z = np.asfortranarray(cdem.z)
    fdem.cellsize = 13.0

    return [cdem, fdem]


def test_flowobject(wide_dem):
    dem = wide_dem
    original_dem = dem.z.copy()
    fd = topo.FlowObject(dem)

    assert topo.validate_alignment(dem, fd)

    # Ensure that the source and target arrays contain valid pixel indices
    assert np.all((0 <= fd.source) & (fd.source < np.prod(dem.shape)))
    assert np.all((0 <= fd.target) & (fd.target < np.prod(dem.shape)))

    # Run flow_accumulation at least once during the tests
    acc = fd.flow_accumulation()

    assert topo.validate_alignment(fd, acc)

    # Ensure that FlowObject does not modify the original DEM
    assert np.all(dem.z == original_dem)


def test_flowobject_order(order_dems):
    cdem, fdem = order_dems

    cfd = topo.FlowObject(cdem)
    ffd = topo.FlowObject(fdem)

    assert isequivalent(cfd, ffd)


def test_flow_accumulation_order(order_dems):
    cdem, fdem = order_dems

    cfd = topo.FlowObject(cdem)
    ffd = topo.FlowObject(fdem)

    ca = cfd.flow_accumulation()
    fa = ffd.flow_accumulation()

    assert np.array_equal(ca, fa)


def test_drainagebasins_order(order_dems):
    cdem, fdem = order_dems

    cfd = topo.FlowObject(cdem)
    ffd = topo.FlowObject(fdem)

    cdb = cfd.drainagebasins()
    fdb = ffd.drainagebasins()

    # This testing logic is rather complicated, so please excuse the
    # length explanation.

    # The two sets of labels are identical up to a bijection between
    # the label sets. We construct that bijection here. We need the
    # set of unique labels, `us`, the indices of a representative pixel
    # in the drainage basins array that has each label, `idxs`, and, for
    # each pixel, the index of its label in the label set, `invs`. If we
    # have B basins in an N x M DEM, `us` is a length B array of basin
    # labels, `idxs` is an length B array of pixel indices and `invs` is
    # an N x M array of the indices [0,B-1) into the `us`
    # array. Fortunately np.unique can provide all of these.
    cus, cidxs, cinvs = np.unique(cdb, return_index=True, return_inverse=True)
    fus, fidxs, finvs = np.unique(fdb, return_index=True, return_inverse=True)

    # Now we construct the bijection and apply it to reconstruct the
    # column-major and row-major drainage basins. Below the "main array"
    # refers to the array whose pattern we are trying to reconstruct
    # and the "alternate array" is the one we reconstruct from.

    # First, we use the `idxs` of the alternate array to extract the
    # `invs` of the main array at each of the representative
    # pixels. Then we index into the `us` of the main array to find
    # the labels assigned to each of the representative pixels /in the
    # order in which they show up in the alternate `idxs` array/,
    # which is also the order of the labels in the alternate `us`
    # array. This array represents the bijection between the main
    # labels and the alternate labels. When we index into this length
    # B array with the alternate `invs`, we will get the main labels
    # that should be assigned to each pixel in the alternate array if
    # the two arrays are identical up to the bijection.
    frec = fus[np.take(finvs, cidxs)][cinvs]

    # Now we test that the reconstructed array is identical to the
    # main array. We use flatten because older versions of numpy
    # return different `invs` of different dimensions.
    assert np.array_equal(frec.flatten(), fdb.z.flatten())

    # Having done this once, we might as well do it with the main and
    # alternate arrays swapped.
    crec = cus[np.take(cinvs, fidxs)][finvs]
    assert np.array_equal(crec.flatten(), cdb.z.flatten())


def test_ezgetnal(wide_dem):
    fd = topo.FlowObject(wide_dem)

    z = fd.ezgetnal(wide_dem)
    z2 = fd.ezgetnal(z.z)
    z3 = fd.ezgetnal(wide_dem, dtype=np.float64)

    assert isinstance(z, topo.GridObject)
    assert isinstance(z2, np.ndarray)
    assert isinstance(z3, topo.GridObject)

    # ezgetnal is idempotent
    assert np.array_equal(z, wide_dem)
    assert np.array_equal(z.z, z2)
    assert np.array_equal(z3, wide_dem)

    # ezgetnal should always return a copy
    assert z is not wide_dem
    assert z2 is not z.z
    assert z3 is not wide_dem

    # ezgetnal with dtype should return array of that dtype
    assert z3.z.dtype is np.dtype(np.float64)


def test_flowpathextract(wide_dem):
    fd = topo.FlowObject(wide_dem)
    s = topo.StreamObject(fd)

    assert topo.validate_alignment(wide_dem, fd)
    assert topo.validate_alignment(wide_dem, s)

    ch = s.streampoi('channelheads')

    for c in s.stream[ch]:

        s2 = topo.StreamObject(fd, channelheads=fd.unravel_index(c))

        assert topo.validate_alignment(wide_dem, s2)

        idxs = fd.flowpathextract(c)

        # NOTE(wkearn): StreamObject's stream nodes are not
        # necessarily topologically sorted, whereas the flow path
        # returned by flowpathextract is topologically sorted. The two
        # arrays will contain the same elements but in different
        # orders. We therefore verify that idxs is the one and only
        # flow path in s2 by reconstructing the topological order of
        # s2.stream from the edge list.
        assert len(idxs) == len(s2.source) + 1
        for e in np.arange(len(s2.source)):
            u = s2.stream[s2.source[e]]
            v = s2.stream[s2.target[e]]
            assert idxs[e] == u
            assert idxs[e + 1] == v

def test_flowpathextract_order(order_dems):
    cdem, fdem = order_dems

    cfd = topo.FlowObject(cdem)
    ffd = topo.FlowObject(fdem)

    # Convert the row-major linear index to column-major
    ci = np.ravel_multi_index((37, 109), cfd.shape, order=cfd.order)
    fi = np.ravel_multi_index((37, 109), ffd.shape, order=ffd.order)

    cp = cfd.flowpathextract(ci)
    fp = ffd.flowpathextract(fi)

    # Convert the column-major flow path indices to row-major
    fpc = np.ravel_multi_index(ffd.unravel_index(fp),
                               cfd.shape, order=cfd.order)

    assert np.array_equal(cp, fpc)

def test_distance_order(order_dems):
    cdem, fdem = order_dems

    cfd = topo.FlowObject(cdem)
    ffd = topo.FlowObject(fdem)

    cd = cfd.node_to_node_distance()
    fd = ffd.node_to_node_distance()

    cdg = np.zeros(cfd.shape)
    fdg = np.zeros(ffd.shape)

    cdg[cfd.source_indices] = cd
    fdg[ffd.source_indices] = fd

    assert np.array_equal(cdg, fdg)

def test_imposemin(wide_dem):
    original_dem = wide_dem.z.copy()
    fd = topo.FlowObject(wide_dem)

    g0 = (wide_dem.z[fd.source_indices] -
          wide_dem.z[fd.target_indices])/fd.node_to_node_distance()

    # Make sure that the test array has slopes less than the imposed minimum
    assert not np.all(g0 >= 0.1 - 1e-6)

    for minimum_slope in [0.0, 0.001, 0.01, 0.1]:
        min_dem = topo.imposemin(fd, wide_dem, minimum_slope)

        # The carved dem should not be above the original
        assert np.all(min_dem.z <= wide_dem)

        # The gradient along the flow network should be greater than or
        # equal to the defined slope within some numerical error
        g = (min_dem.z[fd.source_indices] -
             min_dem.z[fd.target_indices])/fd.node_to_node_distance()
        assert np.all(g >= minimum_slope - 1e-6)

        # imposemin should not modify the original array
        assert np.array_equal(original_dem, wide_dem.z)


def test_imposemin_f64(wide_dem):
    original_dem = np.array(wide_dem, dtype=np.float64)

    fd = topo.FlowObject(wide_dem)

    z = np.array(wide_dem.z, dtype=np.float64)

    min_dem = topo.imposemin(fd, z, 0.001)

    assert np.all(min_dem <= z)

    g = (min_dem[fd.source_indices] -
         min_dem[fd.target_indices])/fd.node_to_node_distance()
    assert np.all(g >= 0.001 - 1e-6)

    # imposemin should not modify the original array
    assert np.array_equal(original_dem, z)

def test_imposemin_order(order_dems):
    cdem, fdem = order_dems

    cfd = topo.FlowObject(cdem)
    ffd = topo.FlowObject(fdem)

    cminslope = topo.imposemin(cfd, cdem, minimum_slope=0.001)
    fminslope = topo.imposemin(ffd, fdem, minimum_slope=0.001)

    assert np.array_equal(cminslope, fminslope)

def test_downstream_distance_order(order_dems):
    cdem, fdem = order_dems

    cfd = topo.FlowObject(cdem)
    ffd = topo.FlowObject(fdem)

    cd = cfd.downstream_distance()
    fd = ffd.downstream_distance()

    assert np.array_equal(cd, fd)

def test_upstream_distance_order(order_dems):
    cdem, fdem = order_dems

    cfd = topo.FlowObject(cdem)
    ffd = topo.FlowObject(fdem)

    cd = cfd.upstream_distance()
    fd = ffd.upstream_distance()

    assert np.array_equal(cd, fd)

def test_dependence_map_order(order_dems):
    cdem, fdem = order_dems
    l_c = cdem.duplicate_with_new_data(np.zeros(cdem.shape, dtype = bool, order = 'C'))
    l_f = fdem.duplicate_with_new_data(np.zeros(fdem.shape, dtype = bool, order = 'F'))
    l_c.z[50:100,50:100] = True
    l_f.z[50:100,50:100] = True

    cfd = topo.FlowObject(cdem)
    ffd = topo.FlowObject(fdem)

    cd = cfd.dependencemap(l_c)
    fd = ffd.dependencemap(l_f)

    assert np.array_equal(cd, fd)

def test_influence_map_order(order_dems):
    cdem, fdem = order_dems
    l_c = cdem.duplicate_with_new_data(np.zeros(cdem.shape, dtype = bool, order = 'C'))
    l_f = fdem.duplicate_with_new_data(np.zeros(fdem.shape, dtype = bool, order = 'F'))
    l_c.z[50:100,50:100] = True
    l_f.z[50:100,50:100] = True

    cfd = topo.FlowObject(cdem)
    ffd = topo.FlowObject(fdem)

    cd = cfd.influencemap(l_c)
    fd = ffd.influencemap(l_f)

    assert np.array_equal(cd, fd)

def test_getoutlets(wide_dem):
    # The outlets determined by FlowObject.getoutlets should be a
    # superset of those determined by s.streampoi('outlets')

    fd = topo.FlowObject(wide_dem)
    s = topo.StreamObject(fd, threshold=1)

    soutlets = s.streampoi('outlets')
    sog = np.zeros(fd.shape, dtype=bool)
    sog[s.node_indices_where(soutlets)] = True

    foutlets = fd.getoutlets()
    fog = np.zeros(fd.shape, dtype=bool)
    fog[fd.unravel_index(foutlets)] = True

    assert np.all(fog[sog])

def test_getoutlets_order(order_dems):
    # Identified outlets should be invariant of the memory ordering.
    cdem, fdem = order_dems

    cfd = topo.FlowObject(cdem)
    ffd = topo.FlowObject(fdem)

    coutlets = cfd.getoutlets()
    cog = np.zeros(cfd.shape, dtype=bool)
    cog[cfd.unravel_index(coutlets)] = True

    foutlets = ffd.getoutlets()
    fog = np.zeros(ffd.shape, dtype=bool)
    fog[ffd.unravel_index(foutlets)] = True

    assert np.count_nonzero(cog) > 0
    assert np.array_equal(cog, fog)

def test_vertdistance2stream_order(order_dems):
    cdem, fdem = order_dems

    cfd = topo.FlowObject(cdem)
    ffd = topo.FlowObject(fdem)

    cs = topo.StreamObject(cfd)
    fs = topo.StreamObject(ffd)

    cv = cfd.vertdistance2stream(cs, cdem)
    fv = ffd.vertdistance2stream(fs, fdem)

    assert np.array_equal(cv, fv)

    cv2 = cfd.vertdistance2stream(cs, fdem)
    fv2 = ffd.vertdistance2stream(fs, cdem)

    assert np.array_equal(cv2, fv2)



import tempfile
import os

import topotoolbox as tt3


def test_write_tif_georef():
    dem = tt3.load_dem('bigtujunga')
    with tempfile.TemporaryDirectory() as tmp:
        path = os.path.join(tmp, 'output.tif')
        tt3.write_tif(dem, path)
        dem2 = tt3.read_tif(path)
        assert dem2.georef == dem.georef

        os.remove(path)



"""Generic TopoToolbox interface

This module contains functions that can be used with all TopoToolbox
objects.
"""

__all__ = ["validate_alignment"]


def validate_alignment(s1, s2) -> bool:
    """Check whether two TopoToolbox objects are aligned

    `validate_alignment` checks that the two objects have the same
    `shape` attribute and, if coordinate information is available, the
    same coordinate system given by the attributes `bounds`, `crs`
    and `transform`.

    Parameters
    ----------
    s1 : np.ndarray | GridObject | FlowObject | StreamObject
        The first object to check

    s2 : np.ndarray | GridObject | FlowObject | StreamObject
        The second object to check

    Returns
    -------
    bool
       True if the two objects are aligned, False otherwise

    Example
    -------
    >>> dem = topotoolbox.load_dem('bigtujunga')
    >>> fd = topotoolbox.FlowObject(dem)
    >>> print(topotoolbox.validate_alignment(dem, fd))
    """
    return (s1.shape == s2.shape) and all(
        (not hasattr(s1, attr) or not hasattr(s2, attr))
        or (getattr(s1, attr) == getattr(s2, attr))
        for attr in ["bounds", "georef", "transform"])



"""This module contains the FlowObject class.
"""
from typing import Literal

import numpy as np

# pylint: disable=no-name-in-module
from . import _grid  # type: ignore
from . import _stream  # type: ignore
from .grid_object import GridObject
from .interface import validate_alignment

__all__ = ['FlowObject']


class FlowObject():
    """A class containing containing (water-) flow information about a given
    digital elevation model (DEM).
    """

    def __init__(self, grid: GridObject,
                 bc: np.ndarray | GridObject | None = None,
                 hybrid: bool = True):
        """The constructor for the FlowObject. Takes a GridObject as input,
        computes flow direction information and saves them as an FlowObject.

        Parameters
        ----------
        grid : GridObject
            The GridObject that will be the basis of the computation.
        bc : ndarray or GridObject, optional
            Boundary conditions for sink filling. `bc` should be an array
            of np.uint8 that matches the shape of the DEM. Values of 1
            indicate pixels that should be fixed to their values in the
            original DEM and values of 0 indicate pixels that should be
            filled.
        hybrid: bool, optional
            Should hybrid reconstruction algorithm be used to fill
            sinks? Defaults to True. Hybrid reconstruction is faster
            but requires additional memory be allocated for a queue.

        Notes
        -----
        Large intermediate arrays are created during the initialization
        process, which could lead to issues when using very large DEMs.

        Example
        -------
        >>> import topotoolbox
        >>> dem = topotoolbox.load_dem('perfectworld')
        >>> flow = topotoolbox.FlowObject(dem)
        """
        dims = grid.dims
        dem = np.asarray(grid, dtype=np.float32)

        filled_dem = np.zeros_like(dem, dtype=np.float32)
        restore_nans = False
        if bc is None:
            bc = np.ones_like(dem, dtype=np.uint8)
            bc[1:-1, 1:-1] = 0  # Set interior pixels to 0

            nans = np.isnan(dem)
            dem[nans] = -np.inf
            bc[nans] = 1
            restore_nans = True

        if not validate_alignment(grid, bc):
            err = ("The shape of the provided boundary conditions does not "
                   f"match the shape of the DEM. {dims}")
            raise ValueError(err)from None

        bc = np.asarray(bc, dtype=np.uint8)

        queue = np.zeros(np.prod(dem.shape), dtype=np.int64)
        if hybrid:
            _grid.fillsinks_hybrid(filled_dem, queue, dem, bc, dims)
        else:
            _grid.fillsinks(filled_dem, dem, bc, dims)

        if restore_nans:
            dem[nans] = np.nan
            filled_dem[nans] = np.nan

        flats = np.zeros_like(dem, dtype=np.int32)
        _grid.identifyflats(flats, filled_dem, dims)

        costs = np.zeros_like(dem, dtype=np.float32)
        conncomps = np.zeros_like(dem, dtype=np.int64)
        _grid.gwdt_computecosts(costs, conncomps, flats, dem, filled_dem, dims)

        dist = np.zeros_like(flats, dtype=np.float32)
        prev = conncomps  # prev: dtype=np.int64
        heap = queue      # heap: dtype=np.int64
        back = np.zeros_like(flats, dtype=np.int64)
        _grid.gwdt(dist, prev, costs, flats, heap, back, dims)

        node = heap  # node: dtype=np.int64
        direction = np.zeros_like(dem, dtype=np.uint8)
        _grid.flow_routing_d8_carve(
            node, direction, filled_dem, dist, flats, dims)

        # ravel is used here to flatten the arrays. The memory order should not matter
        # because we only need a block of contiguous memory interpreted as a 1D array.
        source = np.ravel(conncomps)  # source: dtype=int64
        target = np.ravel(back)       # target: dtype=int64
        edge_count = _grid.flow_routing_d8_edgelist(
            source, target, node, direction, dims)

        self.path = grid.path
        self.name = grid.name

        # raster metadata
        self.direction = direction  # dtype=np.unit8

        self.stream = node
        self.source = source[0:edge_count]  # dtype=np.int64
        self.target = target[0:edge_count]  # dtype=np.int64

        self.shape = grid.shape
        self.cellsize = grid.cellsize
        self.strides = tuple(s // grid.z.itemsize for s in grid.z.strides)
        self.order: Literal['F', 'C'] = ('F' if grid.z.flags.f_contiguous
                                         else 'C')

        # georeference
        self.bounds = grid.bounds
        self.transform = grid.transform
        self.georef = grid.georef

    @property
    def dims(self):
        """The dimensions of the grid in the correct order for libtopotoolbox
        """
        if self.order == 'C':
            return (self.shape[0], self.shape[1])

        return (self.shape[1], self.shape[0])

    @property
    def source_indices(self) -> tuple[np.ndarray, ...]:
        """The row and column indices of the sources of each edge in
        the flow network.

        Returns
        -------
        tuple of ndarray
            A tuple of arrays containing the row indices and column
        indices of the sources of each edge in the flow
        network. Each of these arrays is an edge attribute lists and
        have a length equal to the number of edges in the flow
        network. This tuple of arrays is suitable for indexing
        GridObjects or arrays shaped like the GridObject from which
        this FlowObject was derived.

        """
        return np.unravel_index(self.source, self.shape, self.order)

    @property
    def target_indices(self) -> tuple[np.ndarray, ...]:
        """The row and column indices of the targets of each edge in
        the flow network.

        Returns
        -------
        tuple of ndarray
            A tuple of arrays containing the row indices and column
        indices of the sources of each edge in the flow
        network. Each of these arrays is an edge attribute lists and
        have a length equal to the number of edges in the flow
        network. This tuple of arrays is suitable for indexing
        GridObjects or arrays shaped like the GridObject from which
        this FlowObject was derived.

        """
        return np.unravel_index(self.target, self.shape, self.order)

    def unravel_index(self, idxs: int | np.ndarray) -> tuple[np.ndarray, ...]:
        """Unravel the provided linear indices so they can be used to
        index grids.

        Returns
        -------
        tuple of ndarray
            A tuple of arrays containing the row indices and column
        indices of the sources of each pixel in the idxs array.
        """
        return np.unravel_index(idxs, self.shape, self.order)

    def ezgetnal(self, k, dtype=None) -> GridObject | np.ndarray:
        """Retrieve a node attribute list

        Parameters
        ----------
        k : GridObject or np.ndarray or scalar
            The object from which node values will be extracted. If
            `k` is a `GridObject` or an `ndarray` with the same shape
            as this `FlowObject`, then a copy is returned. If it is a
            scalar, an `ndarray` with the appropriate shape, filled
            with `k`, is returned.

        Returns
        -------
        GridObject or np.ndarray

        Raises
        ------
        ValueError
            The supplied input is not aligned with the FlowObject.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('bigtujunga')
           >>> fd = topotoolbox.FlowObject(dem)
           >>> _= fd.ezgetnal(dem).plot()
           >>> plt.show()
        """
        if np.isscalar(k):
            return np.full(self.shape, k, dtype=dtype)
        if not validate_alignment(self, k):
            raise ValueError("Input is not properly aligned to the FlowObject")

        return k.astype(dtype)

    def flow_accumulation(self, weights: np.ndarray | float = 1.0):
        """Computes the flow accumulation for a given flow network using
        optional weights. The flow accumulation represents the amount of flow
        each cell receives from its upstream neighbors.

        Parameters
        ----------
        weights : np.ndarray | float, optional
            An array of the same shape as the flow grid representing weights
            for each cell, or a constant float value used as the weight for all
            cells. If `weights=1.0` (default), the flow accumulation is
            unweighted. If an ndarray is provided, it must match the shape of
            the flow grid., by default 1.0

        Returns
        -------
        GridObject
            A new GridObject containing the flow accumulation grid.

        Raises
        ------
        ValueError
            If the shape of the `weights` array does not match the shape of the
            flow network grid.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('perfectworld')
           >>> fd = topotoolbox.FlowObject(dem)
           >>> acc = fd.flow_accumulation()
           >>> _= acc.plot(cmap='Blues',norm="log")
           >>> plt.show()
        """
        acc = np.array(self.ezgetnal(weights, dtype=np.float32), copy=True, order=self.order)

        fraction = np.ones_like(self.source, dtype=np.float32)

        _stream.traverse_down_f32_add_mul(acc, fraction, self.source, self.target)

        result = GridObject()
        result.path = self.path
        result.name = self.name

        result.z = acc
        result.cellsize = self.cellsize

        result.bounds = self.bounds
        result.transform = self.transform
        result.georef = self.georef

        return result

    def getoutlets(self):
        """Extract outlets from a flow network.

        These are defined as nodes that have no downstream neighbor in
        the network. This includes any internal sinks as well as
        outlets where flow exits the DEM.

        Returns
        -------
        ndarray
            A list of linear indices of outlets. Use
            `FlowObject.unravel_index` to convert to multidimensional
            indices to index into a `GridObject`.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> from matplotlib.colors import ListedColormap
           >>> dem = topotoolbox.load_dem("bigtujunga")
           >>> fd = topotoolbox.FlowObject(dem)
           >>> outlets = fd.getoutlets()
           >>> j, i = fd.unravel_index(outlets)
           >>> x, y = fd.transform * (i, j)
           >>> _ = dem.plot_hs(cmap=ListedColormap([0.9, 0.9, 0.9]), exaggerate=3)
           >>> _ = plt.scatter(x, y)
           >>> plt.show()
        """
        indegree = np.zeros(self.shape, order=self.order, dtype=np.uint8)
        outdegree = np.zeros(self.shape, order=self.order, dtype=np.uint8)
        _stream.edgelist_degree(indegree, outdegree, self.source, self.target)
        output = (outdegree == 0) & (indegree > 0)

        return np.nonzero(np.ravel(output, order='K'))[0]

    def drainagebasins(self, outlets=None):
        """Delineate drainage basins from a flow network.

        Parameters
        ----------
        outlets: np.ndarray
            An array containing the linear indices of the outlet nodes.

        Returns
        -------
        GridObject
            An integer-valued GridObject with a unique label for each drainage
            basin.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('perfectworld')
           >>> fd = topotoolbox.FlowObject(dem)
           >>> basins = fd.drainagebasins()
           >>> _= basins.shufflelabel().plot(cmap="Pastel1",interpolation="nearest")
           >>> plt.show()
        """
        if outlets is None:
            outlets = self.getoutlets()

        basins = np.zeros(self.shape, dtype=np.uint32, order=self.order)
        indices = self.unravel_index(outlets)
        basins[indices] = np.arange(1, len(outlets) + 1, dtype=np.uint32)
        weights = np.full(self.source.size, 0xffffffff, dtype=np.uint32)
        _stream.traverse_up_u32_or_and(
            basins, weights, self.source, self.target)

        result = GridObject()
        result.path = self.path
        result.name = self.name

        result.z = np.array(basins, dtype=np.int64)
        result.cellsize = self.cellsize

        result.bounds = self.bounds
        result.transform = self.transform
        result.georef = self.georef

        return result

    def flowpathextract(self, idx: int):
        """Extract linear indices of a single flowpath in a DEM

        The flow path downstream of idx is extracted from the flow
        directions recorded in FlowObject.

        Parameters
        ----------
        idx: int
            The column-major linear index of the starting pixel of the flowpath.

        Returns
        -------
        np.ndarray
            An array containing column-major linear indices into the
            DEM identifying the flow path.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> dem = topotoolbox.load_dem('bigtujunga')
           >>> fd = topotoolbox.FlowObject(dem)
           >>> print(fd.flowpathextract(12345)) # doctest: +SKIP
        """
        ch = np.zeros(self.shape, dtype=np.uint32, order=self.order)
        ch[self.unravel_index(idx)] = 1
        edges = np.ones(self.source.size, dtype=np.uint32)
        _stream.traverse_down_u32_or_and(ch, edges, self.source, self.target)

        return self.stream[ch[self.unravel_index(self.stream)] > 0]

    def node_to_node_distance(self):
        """Compute the distance between each node in the flow network

        Returns
        -------
        np.ndarray
            An array (edge attribute list) with the interpixel
            distance. This will be either cellsize or sqrt(2)*cellsize

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> dem = topotoolbox.load_dem('bigtujunga')
           >>> fd = topotoolbox.FlowObject(dem)
           >>> print(fd.node_to_node_distance()) # doctest: +SKIP
        """
        d = np.abs(self.source - self.target)
        dist = self.cellsize * np.where(
            (d == self.strides[0]) | (d == self.strides[1]),
            np.float32(1.0),
            np.sqrt(np.float32(2.0)))
        return dist

    def downstream_distance(self) -> GridObject:
        """Calculates the horizontal distance from outlets and ridges
        along the flow network in the downstream direction.

        Returns
        -------
        down_d: GridObject
            A new GridObject containing the distance grid
        """
        # Edge attribute list
        dist = self.node_to_node_distance()

        down_d = np.zeros(self.shape, dtype = np.float32, order=self.order)
        _stream.traverse_down_f32_max_add(down_d, dist, self.source, self.target)

        result = GridObject()
        result.path = self.path
        result.name = self.name

        result.z = down_d
        result.cellsize = self.cellsize

        result.bounds = self.bounds
        result.transform = self.transform
        result.georef = self.georef

        return result

    def upstream_distance(self) -> GridObject:
        """Calculates the horizontal distance from outlets and ridges
        along the flow network in the upstream direction.

        Returns
        -------
        up_d: GridObject
            A new GridObject containing the distance grid
        """
        # Edge attribute list
        dist = self.node_to_node_distance()

        up_d = np.zeros(self.shape, dtype = np.float32, order=self.order)
        _stream.traverse_up_f32_max_add(up_d, dist, self.source, self.target)

        result = GridObject()
        result.path = self.path
        result.name = self.name

        result.z = up_d
        result.cellsize = self.cellsize

        result.bounds = self.bounds
        result.transform = self.transform
        result.georef = self.georef

        return result

    def vertdistance2stream(self, stream,
                                      grid: GridObject) -> GridObject:
        """Calculates relative elevation from rivers defined by a stream object.
        Follows the flow paths, not the shortest euclidian distance.
        Also called hand (height above nearest drainage)

        Returns
        -------
        up_z: GridObject
            A new GridObject containing the relative elevation grid
        """

        # pylint: disable=import-outside-toplevel
        # Local import to avoid circular import
        from .stream_object import StreamObject

        if not isinstance(stream, StreamObject):
            raise TypeError('stream must be a StreamObject')

        # Getting hte river location as 2D mask
        mask = stream.gridmask

        # Calculating the relative Z from nodes to their receivers

        ## Elevation for every sources and targets
        zsources = grid.z[self.source_indices]
        ztarget = grid.z[self.target_indices]

        ## dz host the local relief
        dz = zsources - ztarget

        # Masking my rivers (if the source node belongs to the river mask)
        dz[mask[self.source_indices]] = 0.

        ## Summing the global relative elevation to rivers
        up_z = np.zeros(self.shape, dtype = np.float32, order=self.order)
        _stream.traverse_up_f32_max_add(up_z, dz, self.source, self.target)

        return grid.duplicate_with_new_data(up_z)

    def dependencemap(self, l) -> GridObject:
        """Delineate upslope area for specific locations in a DEM.

        Parameters
        -------
        fd: FlowObject
        l: GridObject
            logical grid

        Returns
        -------
        i: GridObject
            logical influence grid (GRIDobj)
        """

        # convert input argument to correct units
        seed = self.ezgetnal(l, dtype = np.uint32)

        # graph traversal algorithm
        i = np.ones(self.source.shape, dtype = np.uint32, order=self.order) # turns on all edges
        _stream.traverse_up_u32_or_and(seed, i, self.source, self.target)

        return l.duplicate_with_new_data(np.asarray(seed))

    def influencemap(self, l) -> GridObject:
        """Delineate downslope area for specific locations in a DEM.

        Parameters
        -------
        fd: FlowObject
        l: GridObject
            logical grid

        Returns
        -------
        i: GridObject
            logical influence grid (GRIDobj)
        """

        # convert input argument to correct units
        seed = self.ezgetnal(l, dtype = np.uint32)

        # graph traversal algorithm
        i = np.ones(self.source.shape, dtype = np.uint32, order=self.order) # turns on all edges
        _stream.traverse_down_u32_or_and(seed, i, self.source, self.target)

        return l.duplicate_with_new_data(np.asarray(seed))

    # 'Magic' functions:
    # ------------------------------------------------------------------------

    def __len__(self):
        return len(self.target)

    def __iter__(self):
        return iter(self.target)

    def __getitem__(self, index):
        if isinstance(index, tuple):
            array_name, idx = index

            if array_name == 'target':
                return self.target[idx]
            if array_name == 'source':
                return self.source[idx]
            if array_name == 'direction':
                return self.direction[idx]
            raise ValueError(
                "Invalid raster name ('target', 'source', or 'direction').")
        raise ValueError(
            "Index must be a tuple with (raster_name, index).")

    def __setitem__(self, index, value):
        # Check if the index is a tuple
        if isinstance(index, tuple):
            array_name, idx = index

            if array_name == 'target':
                self.target[idx] = value
            elif array_name == 'source':
                self.source[idx] = value
            elif array_name == 'direction':
                self.direction[idx] = value
            else:
                raise ValueError(
                    "Invalid raster name.('target', 'source', or 'direction')")
        else:
            raise ValueError(
                "Index must be a tuple with (raster_name, index).")

    def __array__(self):
        # Concatenate the arrays along their first axis.
        # Not practical to compute with, but if there is a need to manually
        # plot a FlowObject it'll show the logic nicely.
        return np.concatenate((self.target, self.source, self.direction))
